import re

from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible

from edc_constants.constants import DECLINED, MALE, NEG, NO, POS, UNKNOWN, YES


@deconstructible
class SubjectTypeValidator:
    def __init__(self, subject_types):
        self.subject_types = subject_types

    def __call__(self, value):
        if value not in self.subject_types:
            raise ValidationError(
                "Undefined subject type. Expected one of '{subject_types}'. " "Got '{value}'.",
                params={
                    "subject_types": "' or '".join(self.subject_types),
                    "value": value,
                },
            )

    def __eq__(self, other):
        return self.subject_types == other.subject_types


@deconstructible
class FullNameValidator:
    def __init__(self, regex=None):
        self.regex = regex or re.compile(r"^[A-Z]{1,50}, [A-Z]{1,50}$")

    def __call__(self, value):
        if not re.match(self.regex, value):
            raise ValidationError(
                "Invalid format. Format is 'LASTNAME, FIRSTNAME'. "
                "All uppercase separated by a comma. Note the space "
                "following the comma."
            )

    def __eq__(self, other):
        return self.regex == other.regex


def eligible_if_yes(value):
    if value != YES:
        raise ValidationError("Participant is not eligible.")


def eligible_if_yes_or_declined(value):
    if value not in [YES, DECLINED]:
        raise ValidationError("Please provide the subject with a copy of the consent.")


def eligible_if_no(value):
    if value != NO:
        raise ValidationError("Participant is not eligible.")


def eligible_if_unknown(value):
    if value != UNKNOWN:
        raise ValidationError("Participant is not eligible.")


def eligible_if_female(value):
    if value != "F":
        raise ValidationError("Expected 'Female', Participant is not eligible.")


def eligible_if_male(value):
    if value != MALE:
        raise ValidationError("Expected 'Male', Participant is not eligible.")


def eligible_if_negative(value):
    if value != NEG:
        raise ValidationError(
            "Participant must be HIV Negative." "Participant is not eligible."
        )


def eligible_if_positive(value):
    if value != POS:
        raise ValidationError(
            "Participant must be HIV Positive." "Participant is not eligible."
        )


def eligible_not_positive(value):
    if value == POS:
        raise ValidationError(
            "Participant must be HIV Negative / Unknown." "Participant is not eligible."
        )



from django.contrib.admin import SimpleListFilter
from django.contrib.sites.models import Site

from ..site import sites

__all__ = ["SiteListFilter"]


class SiteListFilter(SimpleListFilter):
    title = "Site"
    parameter_name = "site"

    def lookups(self, request, model_admin):
        names = []
        if model_admin.has_viewallsites_permission(request):
            site_ids = [s.id for s in request.user.userprofile.sites.all()]
        else:
            site_ids = sites.get_site_ids_for_user(request=request)
        for site in Site.objects.filter(id__in=site_ids).order_by("id"):
            names.append((site.id, f"{site.id} {sites.get(site.id).description}"))  # noqa: PERF401
        return tuple(names)

    def queryset(self, request, queryset):  # noqa: ARG002
        if self.value() and self.value() != "none":
            queryset = queryset.filter(site__id=self.value())
        return queryset



from django.contrib.sites.models import Site
from django.db import models
from django_db_views.db_view import DBView

from ...model_mixins import qa_reports_permissions
from .view_definition import get_view_definition


class QaReportLogSummary(DBView):
    username = models.CharField(max_length=100)
    site = models.ForeignKey(Site, on_delete=models.DO_NOTHING)
    report_model = models.CharField(max_length=100)
    first_accessed = models.DateTimeField()
    last_accessed = models.DateTimeField()
    access_count = models.IntegerField()

    view_definition = get_view_definition()

    class Meta:
        managed = False
        db_table = "qa_report_log_summary_view"
        verbose_name = "QA Report Log Summary"
        verbose_name_plural = "QA Report Log Summary"
        default_permissions = qa_reports_permissions



from django.db import models

from edc_constants.choices import YES_NO, YES_NO_NA
from edc_constants.constants import NOT_APPLICABLE


class CitizenFieldsMixin(models.Model):
    citizen = models.CharField(
        verbose_name="Is the participant a Botswana citizen? ",
        max_length=3,
        choices=YES_NO,
    )

    legal_marriage = models.CharField(
        verbose_name=(
            "If not a citizen, is the participant legally " "married to a Botswana citizen?"
        ),
        max_length=3,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
        null=True,
        blank=False,
        help_text="If 'No', participant may not be consented.",
    )

    marriage_certificate = models.CharField(
        verbose_name=(
            "[Interviewer] Has the participant produced " "the marriage certificate as proof? "
        ),
        max_length=3,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
        null=True,
        blank=False,
        help_text="If 'No', participant may not be consented.",
    )

    marriage_certificate_no = models.CharField(
        verbose_name="What is the marriage certificate number?",
        max_length=9,
        null=True,
        blank=True,
        help_text="e.g. 000/YYYY",
    )

    class Meta:
        abstract = True



# Generated by Django 6.0 on 2025-05-05 19:11

import _socket
import django.db.models.deletion
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_audit_fields.models.audit_model_mixin
import django_revision.revision_field
from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ReferenceRangeCollection",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=50, unique=True)),
            ],
            options={
                "verbose_name": "Reference Range Collection",
                "verbose_name_plural": "Reference Range Collections",
                "abstract": False,
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "default_manager_name": "objects",
                "indexes": [
                    models.Index(
                        fields=["modified", "created"],
                        name="edc_reporta_modifie_197943_idx",
                    ),
                    models.Index(
                        fields=["user_modified", "user_created"],
                        name="edc_reporta_user_mo_48976f_idx",
                    ),
                ],
            },
        ),
        migrations.CreateModel(
            name="NormalData",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("label", models.CharField(max_length=25)),
                ("description", models.CharField(max_length=255, null=True)),
                ("reference_group", models.CharField(max_length=25, null=True)),
                ("lower", models.FloatField(null=True)),
                ("lower_operator", models.CharField(max_length=15, null=True)),
                ("lower_inclusive", models.BooleanField(default=False)),
                ("lln", models.CharField(default=None, max_length=15, null=True)),
                ("upper", models.FloatField(null=True)),
                ("upper_operator", models.CharField(max_length=2, null=True)),
                ("upper_inclusive", models.BooleanField(default=False)),
                ("uln", models.CharField(default=None, max_length=15, null=True)),
                ("gender", models.CharField(max_length=15)),
                ("units", models.CharField(max_length=15)),
                ("age_units", models.CharField(max_length=15)),
                ("age_lower", models.IntegerField()),
                ("age_lower_operator", models.CharField(max_length=15)),
                ("age_lower_inclusive", models.BooleanField(default=False)),
                ("age_upper", models.IntegerField()),
                ("age_upper_operator", models.CharField(max_length=15)),
                ("age_upper_inclusive", models.BooleanField(default=False)),
                ("fasting", models.BooleanField(default=False)),
                ("phrase", models.CharField(max_length=50, null=True)),
                ("age_phrase", models.CharField(max_length=25, null=True)),
                ("grade", models.IntegerField(null=True)),
                (
                    "reference_range_collection",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_reportable.referencerangecollection",
                    ),
                ),
            ],
            options={
                "verbose_name": "Normal Reference",
                "verbose_name_plural": "Normal References",
            },
        ),
        migrations.CreateModel(
            name="GradingData",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                        verbose_name="Hostname created",
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                        verbose_name="Hostname modified",
                    ),
                ),
                (
                    "device_created",
                    models.CharField(blank=True, max_length=10, verbose_name="Device created"),
                ),
                (
                    "device_modified",
                    models.CharField(
                        blank=True, max_length=10, verbose_name="Device modified"
                    ),
                ),
                (
                    "locale_created",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale created",
                    ),
                ),
                (
                    "locale_modified",
                    models.CharField(
                        blank=True,
                        help_text="Auto-updated by Modeladmin",
                        max_length=10,
                        null=True,
                        verbose_name="Locale modified",
                    ),
                ),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("label", models.CharField(max_length=25)),
                ("description", models.CharField(max_length=255, null=True)),
                ("reference_group", models.CharField(max_length=25, null=True)),
                ("lower", models.FloatField(null=True)),
                ("lower_operator", models.CharField(max_length=15, null=True)),
                ("lower_inclusive", models.BooleanField(default=False)),
                ("lln", models.CharField(default=None, max_length=15, null=True)),
                ("upper", models.FloatField(null=True)),
                ("upper_operator", models.CharField(max_length=2, null=True)),
                ("upper_inclusive", models.BooleanField(default=False)),
                ("uln", models.CharField(default=None, max_length=15, null=True)),
                ("gender", models.CharField(max_length=15)),
                ("units", models.CharField(max_length=15)),
                ("age_units", models.CharField(max_length=15)),
                ("age_lower", models.IntegerField()),
                ("age_lower_operator", models.CharField(max_length=15)),
                ("age_lower_inclusive", models.BooleanField(default=False)),
                ("age_upper", models.IntegerField()),
                ("age_upper_operator", models.CharField(max_length=15)),
                ("age_upper_inclusive", models.BooleanField(default=False)),
                ("fasting", models.BooleanField(default=False)),
                ("phrase", models.CharField(max_length=50, null=True)),
                ("age_phrase", models.CharField(max_length=25, null=True)),
                ("grade", models.IntegerField()),
                (
                    "reference_range_collection",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_reportable.referencerangecollection",
                    ),
                ),
            ],
            options={
                "verbose_name": "Grading Reference",
                "verbose_name_plural": "Grading References",
            },
        ),
    ]



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from edc_action_item.fieldsets import action_fields, action_fieldset_tuple

from .forms import LtfuForm


class LtfuModelAdminMixin:
    form = LtfuForm

    fieldsets = (
        (None, {"fields": ("subject_identifier", "report_datetime")}),
        (
            "Loss to followup",
            {
                "fields": (
                    "last_seen_datetime",
                    "number_consecutive_missed_visits",
                    "last_missed_visit_datetime",
                    "home_visited",
                    "home_visit_detail",
                    "ltfu_category",
                    "ltfu_category_other",
                    "ltfu_date",
                    "comment",
                )
            },
        ),
        action_fieldset_tuple,
        audit_fieldset_tuple,
    )

    list_display = (
        "subject_identifier",
        "dashboard",
        "ltfu_date",
        "number_consecutive_missed_visits",
        "home_visited",
    )

    list_filter = (
        "ltfu_date",
        "last_seen_datetime",
        "last_missed_visit_datetime",
        "number_consecutive_missed_visits",
    )

    radio_fields = {
        "home_visited": admin.VERTICAL,
        "ltfu_category": admin.VERTICAL,
    }

    search_fields = ("subject_identifier", "action_identifier")

    def get_readonly_fields(self, request, obj=None) -> tuple:
        fields = super().get_readonly_fields(request, obj)
        return action_fields + fields



# Generated by Django 5.1.2 on 2024-11-03 17:08

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("django_pylabels", "0003_rename_updated_labelspecification_modified"),
    ]

    operations = [
        migrations.CreateModel(
            name="LabelSpecification",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("django_pylabels.labelspecification",),
        ),
        migrations.CreateModel(
            name="Drawing",
            fields=[
                (
                    "name",
                    models.CharField(
                        help_text="This is the stored value, required",
                        max_length=250,
                        unique=True,
                        verbose_name="Stored value",
                    ),
                ),
                (
                    "plural_name",
                    models.CharField(max_length=250, null=True, verbose_name="Plural name"),
                ),
                (
                    "display_name",
                    models.CharField(
                        help_text="(suggest 40 characters max.)",
                        max_length=250,
                        unique=True,
                        verbose_name="Name",
                    ),
                ),
                (
                    "display_index",
                    models.IntegerField(
                        default=0,
                        help_text="Index to control display order if not alphabetical, not required",
                        verbose_name="display index",
                    ),
                ),
                (
                    "field_name",
                    models.CharField(
                        blank=True,
                        editable=False,
                        help_text="Not required",
                        max_length=25,
                        null=True,
                    ),
                ),
                ("extra_value", models.CharField(max_length=250, null=True)),
                (
                    "version",
                    models.CharField(default="1.0", editable=False, max_length=35),
                ),
                ("id", models.AutoField(primary_key=True, serialize=False)),
            ],
            options={
                "verbose_name": "Drawing callables",
                "verbose_name_plural": "Drawings callables",
                "abstract": False,
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "indexes": [
                    models.Index(
                        fields=["display_index", "display_name"],
                        name="edc_pylabel_display_00eed7_idx",
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="Label",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=25, unique=True)),
                ("drawing_callable_name", models.CharField(max_length=25)),
                (
                    "label_specification",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="django_pylabels.labelspecification",
                    ),
                ),
            ],
            options={
                "verbose_name": "Label",
                "verbose_name_plural": "Labels",
            },
        ),
    ]



# Generated by Django 2.2.3 on 2019-08-26 13:46

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_data_manager", "0004_auto_20190806_1750")]

    operations = [
        migrations.AlterField(
            model_name="historicalqueryrule",
            name="rule_handler_name",
            field=models.CharField(
                choices=[("do_nothing", "Do Nothing"), ("default", "Default")],
                default="default",
                max_length=150,
            ),
        ),
        migrations.AlterField(
            model_name="queryrule",
            name="rule_handler_name",
            field=models.CharField(
                choices=[("do_nothing", "Do Nothing"), ("default", "Default")],
                default="default",
                max_length=150,
            ),
        ),
    ]



from __future__ import annotations

import csv
import sys
from itertools import islice
from pathlib import Path
from pprint import pprint
from uuid import uuid4

from django.core.exceptions import ObjectDoesNotExist
from django.core.management.color import color_style
from tqdm import tqdm

from edc_sites.site import sites as site_sites

from .exceptions import (
    InvalidAssignment,
    RandomizationListAlreadyImported,
    RandomizationListImportError,
)
from .randomization_list_verifier import RandomizationListVerifier

style = color_style()

__all__ = ["RandomizationListImporter"]


class RandomizationListImporter:
    """Imports upon instantiation a formatted randomization CSV file
    into model RandomizationList.

    default CSV file is the project's randomization_list.csv

    name: name of randomizer, e.g. "default"

    To import SIDS from CSV for the first time:

        from edc_randomization.randomization_list_importer import RandomizationListImporter

        RandomizationListImporter(name='default', add=False, dryrun=False)

        # note: if this is not the first time you will get:
        # RandomizationListImportError: Not importing CSV.
        # edc_randomization.randomizationlist model is not empty!

    To add additional sids from CSV without touching existing model instances:

        from edc_randomization.randomization_list_importer import RandomizationListImporter

        RandomizationListImporter(name='default', add=True, dryrun=False)


    Format:
        sid,site_name, assignment, description, orig_site, orig_allocation, orig_desc
        1,gaborone,intervention,single_dose
        2,gaborone,control,two_doses
        ...
    """

    required_csv_fieldnames = ("sid", "assignment", "site_name", "description")
    verifier_cls = RandomizationListVerifier

    def __init__(
        self,
        randomizer_model_cls=None,
        randomizer_name: str | None = None,
        randomizationlist_path: Path | str | None = None,
        assignment_map: dict[str, int] | None = None,
        verbose: bool | None = None,
        overwrite: bool | None = None,
        add: bool | None = None,
        dryrun: bool | None = None,
        username: str | None = None,
        revision: str | None = None,
        sid_count_for_tests: int | None = None,
        skip_verify: bool | None = None,
        extra_csv_fieldnames: tuple[str] | None = None,
        **kwargs,  # noqa: ARG002
    ):
        extra_csv_fieldnames = extra_csv_fieldnames or ()
        self.verify_messages: str | None = None
        self.add = add
        self.overwrite = overwrite
        self.verbose = True if verbose is None else verbose
        self.dryrun = dryrun
        self.revision = revision
        self.user = username
        self.sid_count_for_tests = sid_count_for_tests
        self.skip_verify = skip_verify
        self.randomizer_model_cls = randomizer_model_cls
        self.randomizer_name = randomizer_name
        self.assignment_map = assignment_map
        self.randomizationlist_path: Path = Path(randomizationlist_path).expanduser()
        self.required_csv_fieldnames = (*self.required_csv_fieldnames, *extra_csv_fieldnames)

        if self.dryrun:
            sys.stdout.write(
                style.MIGRATE_HEADING("\n ->> Dry run. No changes will be made.\n")
            )
        if self.verbose and add:
            count = self.randomizer_model_cls.objects.all().count()
            sys.stdout.write(
                style.SUCCESS(
                    f"\n(*) Randolist model has {count} SIDs (count before import).\n"
                )
            )

    def import_list(self, **kwargs) -> tuple[int, Path]:
        """Imports CSV and verifies."""
        self._raise_on_empty_file()
        self._raise_on_invalid_header()
        self._raise_on_already_imported()
        self._raise_on_duplicates()
        if self.verbose:
            sys.stdout.write(
                style.SUCCESS(
                    "\nImport CSV data\n"
                    "  Randomizer:\n"
                    f"    -  Name: {self.randomizer_name}\n"
                    f"    -  Assignments: {self.assignment_map}\n"
                    f"    -  Model: {self.randomizer_model_cls._meta.label_lower}\n"
                    f"    -  Path: {self.randomizationlist_path}\n"
                )
            )
        rec_count = self._import_csv_to_model()
        if not self.skip_verify:
            self.verify_messages = self._verify_data(**kwargs)
            self._summarize_results()
        if self.verbose:
            sys.stdout.write(
                style.SUCCESS("\nDone.------------------------------------------------\n")
            )
        return rec_count, self.randomizationlist_path

    def _summarize_results(self):
        if self.verbose:
            count = self.randomizer_model_cls.objects.all().count()
            msg = (
                f"\n    - Imported {count} SIDs for randomizer "
                f"`{self.randomizer_name}` into model "
                f"`{self.randomizer_model_cls._meta.label_lower}` \n"
                f"      from {self.randomizationlist_path}.\n"
            )
            sys.stdout.write(style.SUCCESS(msg))
            if self.verify_messages:
                sys.stdout.write(style.ERROR("\n    ! Verification failed. "))
            else:
                sys.stdout.write(style.SUCCESS("    - Verified OK. \n"))

    def _raise_on_empty_file(self):
        if self.randomizationlist_path.stat().st_size < 1:
            raise RandomizationListImportError(
                f"File is empty. See {self.randomizer_name}. "
                f"Got {self.randomizationlist_path} (1)."
            )
        index = 0
        with self.randomizationlist_path.open(mode="r") as csvfile:
            reader = csv.DictReader(csvfile)
            for index, _ in enumerate(reader):
                if index == 0:
                    continue
        if index == 0:
            raise RandomizationListImportError(
                f"File is empty. See {self.randomizer_name}. "
                f"Got {self.randomizationlist_path} (2)."
            )

    def _raise_on_invalid_header(self):
        with self.randomizationlist_path.open(mode="r") as csvfile:
            reader = csv.DictReader(csvfile)
            for index, row in enumerate(reader):
                if index == 0:
                    for fieldname in self.required_csv_fieldnames:
                        if fieldname not in row:
                            raise RandomizationListImportError(
                                f"Invalid header. Missing column `{fieldname}`. Got {row}"
                            )
                elif index == 1:
                    if self.dryrun:
                        row_as_dict = {k: v for k, v in row.items()}
                        sys.stdout.write(" -->  First row:\n")
                        sys.stdout.write(f" -->  {list(row_as_dict.keys())}\n")
                        sys.stdout.write(f" -->  {list(row_as_dict.values())}\n")
                        obj = self.randomizer_model_cls(**self.get_import_options(row))
                        pprint(obj.__dict__)  # noqa: T203
                else:
                    break

    def _raise_on_already_imported(self):
        if not self.dryrun:
            if self.overwrite:
                self.randomizer_model_cls.objects.all().delete()
            if self.randomizer_model_cls.objects.all().count() > 0 and not self.add:
                raise RandomizationListAlreadyImported(
                    "Not importing CSV. "
                    f"{self.randomizer_model_cls._meta.label_lower} "
                    "model is not empty!"
                )

    def get_sid_list(self) -> list[int]:
        with self.randomizationlist_path.open(mode="r") as csvfile:
            reader = csv.DictReader(csvfile)
            sids = [int(row["sid"]) for row in reader]
        if len(sids) != len(list(set(sids))):
            raise RandomizationListImportError("Invalid file. Detected duplicate SIDs")
        return sids

    def _raise_on_duplicates(self) -> None:
        self.get_sid_list()

    def _import_csv_to_model(self) -> int:
        """Imports a CSV to populate the "rando" model"""
        objs = []
        rec_count = 0

        if self.sid_count_for_tests is not None:
            sys.stdout.write(
                style.WARNING(
                    "\nNote: Importing a `subset` of the randomization list for tests "
                    f"({self.sid_count_for_tests}).\n"
                )
            )
            sid_count = self.sid_count_for_tests
        else:
            sid_count = len(self.get_sid_list())
        with self.randomizationlist_path.open(mode="r") as f:
            reader = csv.DictReader(f)
            if self.sid_count_for_tests:
                reader = islice(reader, self.sid_count_for_tests)
            all_rows = [{k: v.strip() for k, v in row.items() if k} for row in reader]
            sorted_rows = sorted(
                all_rows, key=lambda row: (row.get("site_name", ""), row.get("sid", ""))
            )
            for row in tqdm(sorted_rows, total=sid_count):
                # if self.sid_count_for_tests and len(objs) == self.sid_count_for_tests:
                #     break
                try:
                    self.randomizer_model_cls.objects.get(sid=row["sid"])
                except ObjectDoesNotExist:
                    opts = self.get_import_options(row)
                    opts.update(self.get_extra_import_options(row))
                    if self.user:
                        opts.update(user_created=self.user)
                    if self.revision:
                        opts.update(revision=self.revision)
                    obj = self.randomizer_model_cls(**opts)
                    objs.append(obj)
            if not self.dryrun:
                self.randomizer_model_cls.objects.bulk_create(objs)
                rec_count = self.randomizer_model_cls.objects.all().count()
                if not sid_count == rec_count:
                    raise RandomizationListImportError(
                        "Incorrect record count on import. "
                        f"Expected {sid_count}. Got {rec_count}."
                    )
            else:
                sys.stdout.write(
                    style.MIGRATE_HEADING(
                        "\n ->> this is a dry run. No changes were saved. **\n"
                    )
                )
        return rec_count

    def _verify_data(self, **kwargs) -> list[str]:
        verifier = self.verifier_cls(
            assignment_map=self.assignment_map,
            randomizationlist_path=self.randomizationlist_path,
            randomizer_model_cls=self.randomizer_model_cls,
            randomizer_name=self.randomizer_name,
            required_csv_fieldnames=self.required_csv_fieldnames,
            **kwargs,
        )
        return verifier.messages

    def get_assignment(self, row: dict, assignment_map: dict[str, int]) -> str:
        """Returns assignment (text) after checking validity."""
        return self.valid_assignment_or_raise(row["assignment"], assignment_map)

    def get_allocation(self, row: dict, assignment_map: dict[str, int]) -> int:
        """Returns an integer allocation for the given
        assignment or raises.
        """
        assignment = self.get_assignment(row, assignment_map)
        return assignment_map.get(assignment)

    def valid_assignment_or_raise(
        self, assignment: str, assignment_map: dict[str, int]
    ) -> str:
        if assignment not in assignment_map:
            raise InvalidAssignment(
                f"Invalid assignment. Expected one of {list(assignment_map.keys())}. "
                f"Got `{assignment}`. "
                f"See randomizer `{self.randomizer_name}` {self!r}. "
            )
        return assignment

    def get_import_options(self, row):
        return dict(
            id=uuid4(),
            sid=row["sid"],
            assignment=self.get_assignment(row, self.assignment_map),
            allocation=str(self.get_allocation(row, self.assignment_map)),
            randomizer_name=self.randomizer_name,
            site_name=self.validate_site_name(row),
            **self.get_extra_import_options(row),
        )

    def get_extra_import_options(self, row):  # noqa: ARG002
        return {}

    @staticmethod
    def get_site_names() -> dict[str, str]:
        """A dict of site names for the target randomizer."""
        return {
            single_site.name: single_site.name for single_site in site_sites.all().values()
        }

    def validate_site_name(self, row) -> str:
        """Returns the site name or raises"""
        try:
            site_name = self.get_site_names()[row["site_name"].lower()]
        except KeyError as e:
            raise RandomizationListImportError(
                f"Invalid site. Got {row['site_name']}. "
                f"Expected one of {self.get_site_names().keys()}"
            ) from e
        return site_name



from __future__ import annotations

from datetime import datetime
from zoneinfo import ZoneInfo

from django.db import models
from django.db.models import PROTECT
from django.utils.translation import gettext_lazy as _

from edc_appointment.utils import validate_date_is_on_clinic_day
from edc_constants.choices import YES_NO
from edc_constants.constants import NO
from edc_facility.utils import get_health_facility_model

__all__ = ["NextAppointmentCrfModelMixin"]


class NextAppointmentCrfModelMixin(models.Model):

    offschedule_today = models.CharField(
        verbose_name=_("Is the subject going off schedule today?"),
        choices=YES_NO,
        default=NO,
        max_length=15,
        help_text=_("If going off schedule today, additional CRFs will be added for today."),
    )

    appt_date = models.DateField(
        verbose_name=_("Next scheduled routine/facility appointment"),
        help_text=_("Should fall on an valid clinic day for this facility"),
        null=True,
        blank=True,
    )

    appt_datetime = models.DateTimeField(null=True, editable=False)

    info_source = models.ForeignKey(
        "edc_appointment.infosources",
        verbose_name=_("What is the source of this appointment date"),
        max_length=15,
        on_delete=PROTECT,
        null=True,
        blank=True,
    )

    # named this way to not conflict with property `visit_schedule`
    # see also edc-crf
    visitschedule = models.ForeignKey(
        "edc_visit_schedule.VisitSchedule",
        on_delete=PROTECT,
        verbose_name=_("Which study visit code is closest to this appointment date"),
        max_length=15,
        null=True,
        blank=True,
        help_text=_(
            "Click SAVE to let the EDC suggest. Once selected, interim appointments will "
            "be flagged as not required/missed."
        ),
    )

    health_facility = models.ForeignKey(
        get_health_facility_model(),
        on_delete=PROTECT,
        null=True,
        blank=True,
    )

    def save(self, *args, **kwargs):
        if self.appt_date:
            self.appt_datetime = datetime(
                self.appt_date.year,
                self.appt_date.month,
                self.appt_date.day,
                7,
                30,
                0,
                tzinfo=ZoneInfo("UTC"),
            )
            validate_date_is_on_clinic_day(
                cleaned_data=dict(
                    appt_date=self.appt_date,
                    report_datetime=self.report_datetime,
                ),
                clinic_days=self.health_facility.clinic_days,
            )

        super().save(*args, **kwargs)

    class Meta:
        abstract = True
        verbose_name = "Next Appointment"
        verbose_name_plural = "Next Appointments"



from edc_auth.site_auths import site_auths
from edc_auth.utils import remove_default_model_permissions_from_edc_permissions


def update_site_auths() -> None:
    site_auths.add_post_update_func(
        "edc_sites", remove_default_model_permissions_from_edc_permissions
    )


update_site_auths()



from django.template.response import TemplateResponse


def edc_handler400(request, exception=None):
    """400 error handler which includes ``request`` in the context.

    Templates: `400.html`
    Context: None
    """

    context = {"request": request, "message": "bad request", "code": 400}
    template_name = "400.html"  # You need to create a 400.html template.
    return TemplateResponse(request, template_name, context, status=400)


def edc_handler403(request, exception=None):
    """403 error handler which includes ``request`` in the context.

    Templates: `403.html`
    Context: None
    """

    context = {"request": request, "message": "permission denied", "code": 403}
    template_name = "403.html"  # You need to create a 403.html template.
    return TemplateResponse(request, template_name, context, status=403)


def edc_handler404(request, exception=None):
    """404 error handler which includes ``request`` in the context.

    Templates: `404.html`
    Context: None
    """

    context = {"request": request, "message": "page not found", "code": 404}
    template_name = "404.html"  # You need to create a 404.html template.
    return TemplateResponse(request, template_name, context, status=404)


def edc_handler500(request):
    """500 error handler which includes ``request`` in the context.

    Templates: `500.html`
    Context: None
    """

    context = {"request": request, "message": "server error", "code": 500}
    template_name = "500.html"  # You need to create a 500.html template.
    return TemplateResponse(request, template_name, context, status=500)



from edc_model.models import BaseUuidModel

from .container_model_mixin import ContainerModelMixin


class GenericContainer(ContainerModelMixin, BaseUuidModel):
    class Meta(ContainerModelMixin.Meta, BaseUuidModel.Meta):
        verbose_name = "Item"
        verbose_name_plural = "Items"



from django import forms

from edc_visit_schedule.exceptions import OffScheduleError
from edc_visit_schedule.utils import (
    off_all_schedules_or_raise,
    offstudy_datetime_after_all_offschedule_datetimes,
)


class OffstudyModelFormMixin:
    """ModelForm mixin for the Offstudy Model."""

    def clean(self):
        cleaned_data = super().clean()
        self.off_all_schedules_or_raise()
        self.offstudy_datetime_after_all_offschedule_datetimes()
        return cleaned_data

    def off_all_schedules_or_raise(self):
        """Raises a ValidationError if this off study form is submitted
        but subject is still on one or more schedules.
        """
        try:
            off_all_schedules_or_raise(subject_identifier=self.get_subject_identifier())
        except OffScheduleError as e:
            raise forms.ValidationError(e)

    def offstudy_datetime_after_all_offschedule_datetimes(self):
        """Raises a ValidationError if any offschedule datetime is after
        this offstudy_datetime.
        """
        offstudy_datetime_after_all_offschedule_datetimes(
            subject_identifier=self.get_subject_identifier(),
            offstudy_datetime=self.cleaned_data.get("offstudy_datetime"),
            exception_cls=forms.ValidationError,
        )



from .convert_units import convert_units
from .get_default_reportable_grades import get_default_reportable_grades
from .get_grade_for_value import get_grade_for_value
from .get_normal_data_or_raise import get_normal_data_or_raise
from .get_reference_range_collection import get_reference_range_collection
from .grading_data_model_cls import grading_data_model_cls
from .grading_exception_model_cls import grading_exception_model_cls
from .in_normal_bounds_or_raise import in_normal_bounds_or_raise
from .load_data import load_all_reference_ranges, load_reference_ranges
from .molecular_weight_model_cls import molecular_weight_model_cls
from .normal_data_model_cls import normal_data_model_cls
from .reference_range_colllection_model_cls import reference_range_colllection_model_cls
from .update_grading_data import update_grading_data
from .update_grading_exceptions import update_grading_exceptions
from .update_normal_data import update_normal_data



def activate_foreign_keys(sender, connection, **kwargs):
    """Enable integrity constraint with sqlite."""
    if connection.vendor == "sqlite":
        cursor = connection.cursor()
        cursor.execute("PRAGMA foreign_keys = ON;")



from __future__ import annotations

from typing import TYPE_CHECKING

from ..constants import DO_NOTHING, NOT_REQUIRED, REQUIRED

if TYPE_CHECKING:
    from edc_visit_tracking.model_mixins import VisitModelMixin as Base

    from ..model_mixins.creates import CreatesMetadataModelMixin
    from .predicate import PF, P

    class RelatedVisitModel(CreatesMetadataModelMixin, Base):
        pass


class RuleLogicError(Exception):
    pass


class Logic:
    """A class that serves as a container for the logic attribute
    of a rule, the predicate, and the outcomes (result) of that
    rule; consequence and alternative.
    """

    valid_results = [REQUIRED, NOT_REQUIRED, DO_NOTHING]

    def __init__(
        self,
        predicate: P | PF | callable = None,
        consequence: str = None,
        alternative: str = None,
        comment: str | None = None,
    ) -> None:
        if not callable(predicate):
            raise RuleLogicError(
                "Predicate must be a callable. For example a "
                'predicate class such as "P" or "PF"'
            )
        self.predicate = predicate
        self.consequence = consequence
        self.alternative = alternative
        self.comment = comment
        for result in [self.consequence, self.alternative]:
            if result not in self.valid_results:
                raise RuleLogicError(
                    f"Invalid result on rule. Expected one of "
                    f"{self.valid_results}. Got {result}."
                )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.predicate}, "
            f"{self.consequence}, {self.alternative})"
        )



from __future__ import annotations

from .deactivate_refill import deactivate_refill
from .get_active_refill import get_active_refill


def activate_refill(rx_refill):
    """Activates this rx_refill and deactivates the active rx_refill,
    if there is one.
    """
    if active_rx_refill := get_active_refill(rx_refill.rx):
        if active_rx_refill != rx_refill:
            deactivate_refill(active_rx_refill)
    rx_refill.active = True
    rx_refill.save()
    rx_refill.refresh_from_db()



from edc_protocol.research_protocol_config import ResearchProtocolConfig

from .base_label import BaseLabel


class BoxLabel(BaseLabel):
    model = "edc_lab.box"
    template_name = "box"

    @property
    def label_context(self):
        return {
            "barcode_value": self.model_obj.box_identifier,
            "box_identifier": self.model_obj.human_readable_identifier,
            "box_name": self.model_obj.name,
            "protocol": ResearchProtocolConfig().protocol,
            "site": str(self.model_obj.site.id),
            "box_datetime": self.model_obj.box_datetime.strftime("%Y-%m-%d %H:%M"),
            "category": self.model_obj.get_category_display().upper(),
            "specimen_types": self.model_obj.specimen_types,
            "site_name": str(self.model_obj.site.siteprofile.title),
        }



from django.db.models import QuerySet, Sum
from django.utils.translation import gettext as _
from reportlab.graphics.barcode import code128
from reportlab.lib import colors
from reportlab.lib.enums import TA_CENTER, TA_LEFT, TA_RIGHT
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.units import cm, mm
from reportlab.pdfbase.pdfmetrics import stringWidth
from reportlab.platypus import Paragraph, SimpleDocTemplate, Spacer, Table, TableStyle

from edc_pdf_reports import Report
from edc_protocol.research_protocol_config import ResearchProtocolConfig

from ..models import Stock
from ..utils import get_related_or_none


class StockReport(Report):

    def __init__(self, queryset: QuerySet[Stock] = None, **kwargs):
        self.queryset = queryset.order_by("from_stock__code", "code")
        self.protocol_name = ResearchProtocolConfig().protocol_title
        super().__init__(**kwargs)

    def draw_header(self, canvas, doc):
        width, height = self.page.get("pagesize")
        canvas.setFontSize(6)
        text_width = stringWidth(self.protocol_name, "Helvetica", 6)
        canvas.drawRightString(width - text_width, height - 20, self.protocol_name.upper())
        canvas.drawString(
            40,
            height - 30,
            _("Stock Report").upper(),
        )

    def get_report_story(self, document_template: SimpleDocTemplate = None, **kwargs):
        story = []

        data = [
            [
                Paragraph(
                    _("Stock Report").upper(),
                    ParagraphStyle(
                        "Title",
                        fontSize=10,
                        spaceAfter=0,
                        alignment=TA_LEFT,
                        fontName="Helvetica-Bold",
                    ),
                ),
                Paragraph(
                    self.protocol_name.upper(),
                    ParagraphStyle(
                        "Title",
                        fontSize=10,
                        spaceAfter=0,
                        alignment=TA_RIGHT,
                        fontName="Helvetica-Bold",
                    ),
                ),
            ],
        ]
        table = Table(data)
        story.append(table)
        story.append(Spacer(0.1 * cm, 0.5 * cm))

        story.append(self.stock_items_as_table)

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        return story

    @property
    def stock_items_as_table(self) -> Table:

        style = ParagraphStyle(
            name="line_data_medium",
            alignment=TA_CENTER,
            fontSize=8,
            textColor=colors.black,
            fontName="Helvetica-Bold",
        )
        data = [
            [
                Paragraph("#", style),
                Paragraph(_("BARCODE"), style),
                Paragraph(_("STOCK"), style),
                Paragraph(_("FROM"), style),
                Paragraph(_("SUBJECT"), style),
                Paragraph(_("FORMULATION"), style),
                Paragraph(_("PACK"), style),
                Paragraph(_("CATSD"), style),
                Paragraph(_("IN"), style),
                Paragraph(_("OUT"), style),
                Paragraph(_("QTY"), style),
            ]
        ]
        for index, stock_obj in enumerate(self.queryset.all()):
            barcode = code128.Code128(stock_obj.code, barHeight=5 * mm, barWidth=0.7, gap=1.7)
            subject_identifier = (
                stock_obj.allocation.registered_subject.subject_identifier
                if get_related_or_none(stock_obj, "allocation")
                else ""
            )
            formulation = stock_obj.product.formulation
            description = f"{formulation.imp_description} "
            catsd = self.get_catsbd(stock_obj)
            style = ParagraphStyle(
                name="line_data", alignment=TA_CENTER, fontSize=8, leading=10
            )
            num_style = ParagraphStyle(
                name="line_data", alignment=TA_RIGHT, fontSize=8, leading=10
            )
            style_xsmall = ParagraphStyle(
                name="line_data", alignment=TA_CENTER, fontSize=6, leading=8
            )
            from_stock_code = ""
            if stock_obj.from_stock:
                from_stock_code = stock_obj.from_stock.code
            qty_in = str(int(stock_obj.unit_qty_in))
            qty_out = str(int(stock_obj.unit_qty_out))
            qty = str(int(stock_obj.unit_qty_in - stock_obj.unit_qty_out))
            data.append(
                [
                    Paragraph(str(index + 1), style),
                    barcode,
                    Paragraph(stock_obj.code, style),
                    Paragraph(from_stock_code, style),
                    Paragraph(subject_identifier, style),
                    Paragraph(description, style_xsmall),
                    Paragraph(str(stock_obj.container), style),
                    Paragraph(
                        catsd,
                        ParagraphStyle(
                            name="line_data",
                            alignment=TA_CENTER,
                            fontName="Courier",
                            fontSize=8,
                            leading=10,
                        ),
                    ),
                    Paragraph(qty_in, num_style),
                    Paragraph(qty_out, num_style),
                    Paragraph(qty, num_style),
                ]
            )
        totals = self.queryset.values("unit_qty_in", "unit_qty_out").aggregate(
            Sum("unit_qty_in"), Sum("unit_qty_out")
        )
        tot_qty_in = totals.get("unit_qty_in__sum")
        tot_qty_out = totals.get("unit_qty_out__sum")
        total = tot_qty_in - tot_qty_out
        data.append(
            [
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                Paragraph(str(int(tot_qty_in)), num_style),
                Paragraph(str(int(tot_qty_out)), num_style),
                Paragraph(str(int(total)), num_style),
            ]
        )

        table = Table(
            data,
            colWidths=(
                1 * cm,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
            ),
        )
        table.setStyle(
            TableStyle(
                [
                    ("BACKGROUND", (0, 0), (-1, 0), colors.white),
                    ("TEXTCOLOR", (0, 0), (-1, 0), colors.black),
                    ("ALIGN", (0, 0), (-1, -1), "CENTER"),
                    ("FONTNAME", (0, 0), (-1, 0), "Helvetica-Bold"),
                    ("BOTTOMPADDING", (0, 0), (-1, 0), 0),
                    ("BACKGROUND", (0, 1), (-1, -1), colors.white),
                    ("GRID", (0, 0), (-1, -1), 1, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        return table

    @staticmethod
    def get_catsbd(stock_obj: Stock) -> str:
        catsd = ""
        catsd += "C" if get_related_or_none(stock_obj, "confirmation") else "-"
        catsd += "A" if get_related_or_none(stock_obj, "allocation") else "-"
        catsd += "T" if get_related_or_none(stock_obj, "stocktransferitem") else "-"
        catsd += "S" if get_related_or_none(stock_obj, "confirmationatsiteitem") else "-"
        catsd += "B" if get_related_or_none(stock_obj, "stored_at_site") else "-"
        catsd += "D" if get_related_or_none(stock_obj, "dispenseitem") else "-"
        return catsd



import numpy as np


class DfHandlerUnexpectedRowCount(Exception):
    pass


class DfHandlerDuplicateColumn(Exception):
    pass


class DfHandler:
    column_handler_cls = None
    na_value = np.nan
    sort_by = None

    def __init__(
        self,
        dataframe=None,
        db=None,
        sort_by=None,
        na_value=None,
        table_name=None,
        **kwargs,
    ):
        self.table_name = table_name
        self.dataframe = dataframe
        original_row_count = len(self.dataframe.index)
        self.na_value = self.na_value if na_value is None else na_value
        self.sort_by = self.sort_by if sort_by is None else sort_by

        self.db = db

        # raise on duplicate column
        cols = list(self.dataframe.columns)
        diff = set([x for x in cols if cols.count(x) > 1])
        if len(diff) > 0:
            raise DfHandlerDuplicateColumn(f"Duplicate column detected. Got {list(diff)}")

        self.prepare_dataframe()
        self.finish_dataframe()

        # verify size has not changed
        if original_row_count != len(self.dataframe.index):
            raise DfHandlerUnexpectedRowCount(
                f"Dataframe count has changed. Expected {original_row_count}. "
                f"Got {len(self.dataframe.index)} "
            )

    def __repr__(self):
        return f"{self.__class__.__name__}(table_name={self.table_name})"

    def prepare_dataframe(self, **kwargs):
        pass

    def finish_dataframe(self, **kwargs):
        if self.column_handler_cls:
            handler = self.column_handler_cls(self.dataframe)
            self.dataframe = handler.dataframe

        for column in list(
            self.dataframe.select_dtypes(include=["datetime64[ns, UTC]"]).columns
        ):
            self.dataframe[column] = self.dataframe[column].astype("datetime64[ns]")
        self.dataframe.fillna(value=self.na_value, inplace=True)
        if self.sort_by:
            self.dataframe.sort_values(by=self.sort_by, inplace=True)



from reportlab.lib import colors
from reportlab.lib.pagesizes import A4
from reportlab.platypus import Flowable


class TextboxFlowable(Flowable):
    def __init__(self, name, width=200, height=60, value="", borderStyle="solid"):
        Flowable.__init__(self)
        self.name = name
        self.width = width
        self.height = height
        self.value = value
        self.borderStyle = borderStyle

    def draw(self):
        # Calculate the position based on the current coordinates
        x = self.canv._currentMatrix[4]
        y = self.canv._currentMatrix[5] - self.height

        # Draw the bottom border if borderStyle is 'underlined'
        if self.borderStyle == "underlined":
            self.canv.setStrokeColor(colors.black)
            self.canv.setLineWidth(1)
            self.canv.line(x, y, x + self.width, y)

        # Create the text area with the specified border style
        width, height = A4
        self.canv.acroForm.textfield(
            name=self.name,
            tooltip=self.name,
            x=x,
            y=y,
            width=width * 0.6,
            height=height * 0.4,
            borderStyle=(self.borderStyle if self.borderStyle != "underlined" else "none"),
            fillColor=colors.white,
            textColor=colors.black,
            forceBorder=True,
            value=self.value,
            fieldFlags="multiline",  # Enable multiline for the textbox
        )

    def split(self, availWidth, availHeight):
        return [self]



from .processing_profile import ProcessingProfile
from .requisition_panel import RequisitionPanel


class RequisitionPanelGroupError(Exception):
    pass


class RequisitionPanelGroup(RequisitionPanel):
    def __init__(
        self,
        *panels,
        name: str | None = None,
        verbose_name: str | None = None,
        abbreviation: str | None = None,
        is_poc: bool | None = None,
        reference_range_collection_name: str | None = None,
    ):
        processing_profile = ProcessingProfile(
            name=f"{name} group", aliquot_type=panels[0].aliquot_type
        )
        utest_ids = []
        for panel in panels:
            utest_ids.extend(panel.utest_ids)
            if (
                panel.reference_range_collection_name
                and reference_range_collection_name != panel.reference_range_collection_name
            ):
                raise RequisitionPanelGroupError(
                    "Panels in a panel group must use the same reference range "
                    f"collection name. Got "
                    f"{reference_range_collection_name} != "
                    f"{panel.reference_range_collection_name}. "
                    f"See {self}."
                )
            if panel.aliquot_type and panel.aliquot_type != processing_profile.aliquot_type:
                raise RequisitionPanelGroupError(
                    "Panels in a panel group must use the same aliquot_type "
                    f"Got {panel.aliquot_type} != "
                    f"{panel.processing_profile.aliquot_type}. "
                    f"See {self}."
                )
        super().__init__(
            utest_ids=utest_ids,
            processing_profile=processing_profile,
            name=name,
            verbose_name=verbose_name,
            abbreviation=abbreviation,
            is_poc=is_poc,
            reference_range_collection_name=reference_range_collection_name,
        )

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    def __str__(self):
        return self.verbose_name or self.name



from .medication import DosageGuidelineForm, FormulationForm, MedicationForm
from .prescription import RxForm, RxRefillForm
from .stock import (
    ConfirmationForm,
    ContainerForm,
    ContainerTypeForm,
    ContainerUnitsForm,
    DispenseForm,
    LocationForm,
    LotForm,
    OrderForm,
    OrderItemForm,
    ProductForm,
    ReceiveForm,
    ReceiveItemForm,
    RepackRequestForm,
    StockForm,
    StockRequestForm,
    StockRequestItemForm,
    StockTransferForm,
    StorageBinForm,
    StorageBinItemForm,
    SupplierForm,
)



from __future__ import annotations

from pathlib import Path
from tempfile import mkdtemp
from typing import TYPE_CHECKING

from django.utils import timezone

from edc_pdutils.df_exporters import CsvModelExporter
from edc_sites.site import sites

from .files_archiver import FilesArchiver
from .files_emailer import FilesEmailer, FilesEmailerError

if TYPE_CHECKING:
    from datetime import datetime

    from django.contrib.auth.models import User


class ArchiveExporterNothingExported(Exception):  # noqa: N818
    pass


class ArchiveExporterEmailError(Exception):
    pass


class ArchiveExporter:
    """Exports a list of models to individual CSV files and
    adds each to a single zip archive OR emails each.

    models: a list of model names in label_lower format.
    """

    date_format: str = "%Y%m%d%H%M%S"
    csv_exporter_cls: type[CsvModelExporter] = CsvModelExporter
    files_emailer_cls: type[FilesEmailer] = FilesEmailer
    files_archiver_cls: type[FilesArchiver] = FilesArchiver

    def __init__(
        self,
        models: list[str] = None,
        decrypt: bool | None = None,
        user: User = None,
        archive: bool | None = None,
        email_to_user: bool | None = None,
        **kwargs,
    ):
        models: list[str] = models or []
        self.archive_filename: str | None = None
        self.exported: list = []
        self.emailed_to: str | None = None
        self.emailed_datetime: datetime | None = None
        self.exported_datetime: datetime | None = None
        tmp_folder: str = mkdtemp()
        for model in models:
            csv_exporter = self.csv_exporter_cls(
                model=model,
                export_folder=Path(tmp_folder),
                decrypt=decrypt,
                site_ids=sites.get_site_ids_for_user(
                    user=user, site_id=sites.get_current_site().site_id
                ),
                **kwargs,
            )
            self.exported.append(csv_exporter.to_csv())
        if not self.exported:
            raise ArchiveExporterNothingExported(f"Nothing exported. Got models={models}.")
        if archive:
            archiver = self.files_archiver_cls(
                path=tmp_folder,
                user=user,
                exported_datetime=self.exported_datetime,
                date_format=self.date_format,
            )
            self.archive_filename = archiver.archive_filename
            self.exported_datetime = archiver.exported_datetime
        if email_to_user:
            summary = [str(x) for x in self.exported]
            summary.sort()
            try:
                self.files_emailer_cls(
                    path=tmp_folder,
                    user=user,
                    file_ext=".zip" if archive else ".csv",
                    summary="\n".join(summary),
                )
            except FilesEmailerError as e:
                raise ArchiveExporterEmailError(e) from e
            else:
                self.emailed_to = user.email
                self.emailed_datetime = timezone.now()
                self.exported_datetime = self.exported_datetime or self.emailed_datetime



import django.utils.timezone
from django.db import models
from sequences import get_next_value

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_sites.model_mixins import SiteModelMixin
from edc_visit_schedule.model_mixins import VisitCodeFieldsModelMixin

from .dispense import Dispense
from .stock import Stock


class Manager(models.Manager):
    use_in_migrations = True


class DispenseItem(SiteModelMixin, VisitCodeFieldsModelMixin, BaseUuidModel):
    """A model for the stock dispensing transaction."""

    dispense_item_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    dispense_item_datetime = models.DateTimeField(default=django.utils.timezone.now)

    dispense = models.ForeignKey(
        Dispense,
        verbose_name="Dispense",
        on_delete=models.PROTECT,
        null=True,
        blank=False,
    )

    stock = models.OneToOneField(Stock, on_delete=models.PROTECT)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return self.dispense_item_identifier

    def save(self, *args, **kwargs):
        self.site = self.dispense.site
        if not self.dispense_item_identifier:
            self.dispense_item_identifier = f"{get_next_value(self._meta.label_lower):010d}"
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Dispense item"
        verbose_name_plural = "Dispense items"



# Generated by Django 4.2.7 on 2023-12-06 16:29

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_lab", "0031_alter_box_options_alter_boxitem_options_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="consignee",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Consignee",
            },
        ),
        migrations.RemoveIndex(
            model_name="consignee",
            name="edc_lab_con_name_2de87e_idx",
        ),
        migrations.AddIndex(
            model_name="consignee",
            index=models.Index(
                fields=["modified", "created"], name="edc_lab_con_modifie_970fb3_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="consignee",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_lab_con_user_mo_a53323_idx",
            ),
        ),
    ]



import sys

from django.apps import apps as django_apps
from django.conf import settings
from django.core.management.color import color_style

style = color_style()


def populate_data_dictionary(sender=None, **kwargs):
    from .populate_data_dictionary import populate_data_dictionary_from_sites

    sys.stdout.write(style.MIGRATE_HEADING("Populating data dictionary:\n"))
    if getattr(settings, "EDC_DATA_MANAGER_POPULATE_DATA_DICTIONARY", True):
        populate_data_dictionary_from_sites()
        sys.stdout.write("Done populating data dictionary.\n\n")
    else:
        sys.stdout.write(
            "  not populating. See settings.EDC_DATA_MANAGER_POPULATE_DATA_DICTIONARY\n\n"
        )
    sys.stdout.flush()


def update_query_rule_handlers(sender=None, **kwargs):
    from .site_data_manager import site_data_manager

    sys.stdout.write(
        style.MIGRATE_HEADING("Deactivating query rules with invalid rule handler names:\n")
    )
    handler_names = [x for x in site_data_manager.registry.keys()]
    django_apps.get_model("edc_data_manager.queryrule").objects.exclude(
        rule_handler_name__in=handler_names
    ).update(active=False)
    sys.stdout.write("Done.\n")
    sys.stdout.flush()



# Generated by Django 3.2 on 2021-05-10 17:36

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_export", "0008_auto_20201002_0403"),
    ]

    operations = [
        migrations.AlterField(
            model_name="exportdata",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="importdata",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
    ]



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Any

from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone

from edc_visit_schedule.models import VisitSchedule

if TYPE_CHECKING:
    from edc_visit_tracking.model_mixins import VisitModelMixin


class SubjectVisitViewMixinError(Exception):
    pass


class SubjectVisitViewMixin:
    """Mixin to add the subject visit instance to the view.

    Declare together with the edc_appointment.AppointmentViewMixin.
    """

    def __init__(self, **kwargs):
        self._related_visit = None
        self._report_datetime = None
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        visit_schedule_pk = (
            "" if not self.get_visit_schedule() else str(self.get_visit_schedule().id)
        )
        kwargs.update(
            related_visit=self.related_visit,
            visit_schedule_model_obj=self.get_visit_schedule(),
            visit_schedule_pk=visit_schedule_pk,
        )
        return super().get_context_data(**kwargs)

    @property
    def related_visit(self) -> VisitModelMixin | None:
        if not self._related_visit:
            self.has_appointment_attr_or_raise()
            if self.appointment:
                try:
                    self._related_visit = self.appointment.related_visit
                except AttributeError as e:
                    raise SubjectVisitViewMixinError(
                        f"Visit model must have a OneToOne relation to appointment. Got {e}"
                    ) from e
        return self._related_visit

    def get_visit_schedule(self) -> VisitSchedule | None:
        """Returns the VisitSchedule model instance."""
        self.has_appointment_attr_or_raise()
        visit_schedule = None
        if self.appointment:
            opts = dict(
                visit_schedule_name=self.appointment.visit_schedule_name,
                schedule_name=self.appointment.schedule_name,
                visit_code=self.appointment.visit_code,
            )
            with contextlib.suppress(ObjectDoesNotExist):
                visit_schedule = VisitSchedule.objects.get(**opts)
        return visit_schedule

    @property
    def report_datetime(self):
        """Returns the report datetime in UTC"""
        if not self._report_datetime:
            try:
                self._report_datetime = self.related_visit.report_datetime
            except AttributeError:
                try:
                    self._report_datetime = self.appointment.appt_datetime
                except AttributeError:
                    self._report_datetime = timezone.now()
        return self._report_datetime

    def has_appointment_attr_or_raise(self) -> None:
        """Checks for 'appointment' attribute. If missing, assume you
        did not declare this view with AppointmentViewMixin.
        """
        try:
            self.appointment  # noqa: B018
        except AttributeError as e:
            if "appointment" in str(e):
                raise SubjectVisitViewMixinError(
                    f"Mixin must be declared together with AppointmentViewMixin. Got {e}"
                ) from e
            raise



from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from django_audit_fields.admin import audit_fieldset_tuple
from rangefilter.filters import DateRangeFilterBuilder

from edc_model_admin.history import SimpleHistoryAdmin
from edc_sites.admin import SiteModelAdminMixin
from edc_utils.date import to_local

from ...admin_site import edc_pharmacy_admin
from ...forms import DispenseForm
from ...models import Dispense
from ..model_admin_mixin import ModelAdminMixin


@admin.register(Dispense, site=edc_pharmacy_admin)
class DispenseAdmin(SiteModelAdminMixin, ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Dispense"
    change_form_title = "Pharmacy: Dispense"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    form = DispenseForm

    fieldsets = (
        (
            "Section A: Dispense",
            {
                "fields": (
                    "dispense_identifier",
                    "dispense_datetime",
                    "location",
                    "rx",
                    "dispensed_by",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "identifier",
        "dispense_date",
        "subject",
        "formulation",
        "location",
        "dispense_item_changelist",
        "stock_changelist",
    )

    search_fields = (
        "id",
        "rx__subject_identifier",
        "dispenseitem__stock__code",
    )

    list_filter = (
        ("dispense_datetime", DateRangeFilterBuilder()),
        "location__display_name",
    )

    readonly_fields = (
        "dispense_identifier",
        "dispense_datetime",
        "rx",
        "location",
        "dispensed_by",
    )

    @admin.display(description="Dispense #", ordering="-dispense_identifier")
    def identifier(self, obj):
        return obj.dispense_identifier

    @admin.display(
        description="Subject #", ordering="rx__registered_subject__subject_identifier"
    )
    def subject(self, obj):
        return obj.rx.registered_subject.subject_identifier

    @admin.display(description="Formulation")
    def formulation(self, obj):
        return "<BR>".join([o.display_name for o in obj.rx.medications.all()])

    @admin.display(description="Dispense date", ordering="dispense_datetime")
    def dispense_date(self, obj):
        return to_local(obj.dispense_datetime).date()

    @admin.display(description="Dispense items")
    def dispense_item_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_dispenseitem_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(url=url, label="Dispense items", title="Go to items")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Stock")
    def stock_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
        url = f"{url}?q={obj.id}"
        context = dict(url=url, label="Stock", title="Go to stock")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    def get_view_only_site_ids_for_user(self, request) -> list[int]:
        return [s.id for s in request.user.userprofile.sites.all() if s.id != request.site.id]



from __future__ import annotations

from typing import TYPE_CHECKING

from ..utils import get_duplicates
from .forms_collection import FormsCollection, FormsCollectionError

if TYPE_CHECKING:
    from .crf import Crf


class CrfCollection(FormsCollection):
    def __init__(self, *forms: Crf, name: str | None = None, **kwargs):
        super().__init__(*forms, name=name, **kwargs)

    @staticmethod
    def collection_is_unique_or_raise(forms: tuple[Crf]):
        models = [f.model for f in forms if f.required]
        if duplicates := get_duplicates(list_items=models):
            raise FormsCollectionError(
                "Expected to be a unique sequence of crf/models. "
                f"Got {sorted(models)}. Duplicates {sorted(duplicates)}."
            )



from django import forms

from ...models import Lot


class LotForm(forms.ModelForm):

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data.get("product") and cleaned_data.get("assignment"):
            if cleaned_data.get("product").assignment != cleaned_data.get("assignment"):
                raise forms.ValidationError(
                    {"assignment": "Assignment does not match product"}
                )

        return cleaned_data

    class Meta:
        model = Lot
        fields = "__all__"



# views.py
from django.db.models import Count
from django.http import JsonResponse

from ..models import StockTransfer


def get_stock_transfers_view(request):
    location_id = request.GET.get("location_id", None)
    stock_transfers = (
        StockTransfer.objects.filter(
            to_location_id=location_id,
            stocktransferitem__stock__confirmationatsiteitem__isnull=True,
        )
        .annotate(count=Count("transfer_identifier"))
        .values("id", "transfer_identifier", "item_count")
        .order_by("-transfer_identifier")
    )
    return JsonResponse(list(stock_transfers), safe=False)



from django.db import models


class NotificationModelMixin(models.Model):
    emailed = models.BooleanField(default=False)

    emailed_datetime = models.DateTimeField(null=True)

    class Meta:
        abstract = True



import sys

from django.apps import AppConfig as DjangoAppConfig
from django.core.checks.registry import register
from django.core.management.color import color_style

from .system_checks import holiday_country_check, holiday_path_check
from .utils import get_facilities

style = color_style()


class AppConfig(DjangoAppConfig):
    _holidays = {}
    name = "edc_facility"
    verbose_name = "Edc Facility"
    include_in_administration_section = True

    def ready(self):
        sys.stdout.write(f"Loading {self.verbose_name} ...\n")
        if "migrate" not in sys.argv and "showmigrations" not in sys.argv:
            register(holiday_path_check, deploy=True)
            register(holiday_country_check, deploy=True)
        else:
            sys.stdout.write(
                style.NOTICE(" * not registering system checks for migrations.\n")
            )
        for facility in get_facilities().values():
            sys.stdout.write(f" * {facility}.\n")
        sys.stdout.write(f" Done loading {self.verbose_name}.\n")



from django import forms

from ...models import Stock


class StockForm(forms.ModelForm):

    class Meta:
        model = Stock
        fields = "__all__"
        help_text = {
            "stock_identifier": "(read-only)",
            "receive_item": "(read-only)",
            "container": "(read-only)",
        }
        widgets = {
            "stock_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
            "receive_item": forms.TextInput(attrs={"readonly": "readonly"}),
            "container": forms.TextInput(attrs={"readonly": "readonly"}),
        }



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING
from uuid import UUID

from django.apps import apps as django_apps

from .crf_button import CrfButton

if TYPE_CHECKING:
    from edc_lab.models import Panel

__all__ = ["RequisitionButton"]


@dataclass
class RequisitionButton(CrfButton):
    @property
    def btn_id(self) -> str:
        btn_id = super().btn_id
        if self.model_obj:
            btn_id = f"{btn_id}-{self.panel.name}"
        return btn_id

    @property
    def extra_kwargs(self) -> dict[str, str | int | UUID]:
        extra_kwargs = super().extra_kwargs
        extra_kwargs.update(panel=self.panel.id)
        return extra_kwargs

    @property
    def panel(self) -> Panel:
        if self.model_obj:
            panel = self.model_obj.panel
        else:
            panel_model_cls = django_apps.get_model("edc_lab.panel")
            panel = panel_model_cls.objects.get(name=self.metadata_model_obj.panel_name)
        return panel



from django.contrib import admin

from edc_list_data.admin import ListModelAdminMixin

from ...admin_site import edc_pharmacy_admin
from ...forms import ContainerUnitsForm
from ...models import ContainerUnits


@admin.register(ContainerUnits, site=edc_pharmacy_admin)
class ContainerUnitsAdmin(ListModelAdminMixin, admin.ModelAdmin):
    show_object_tools = True

    form = ContainerUnitsForm



# Generated by Django 5.0 on 2023-12-08 00:49

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_data_manager", "0035_datadictionary_edc_data_ma_model_92b469_idx"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalqueryrule",
            name="rule_handler_name",
            field=models.CharField(
                choices="(('do_nothing', 'Do Nothing'), ('default', 'Default'))",
                default="default",
                max_length=150,
            ),
        ),
        migrations.AlterField(
            model_name="queryrule",
            name="rule_handler_name",
            field=models.CharField(
                choices="(('do_nothing', 'Do Nothing'), ('default', 'Default'))",
                default="default",
                max_length=150,
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2017-01-06 19:18
from __future__ import unicode_literals

from django.db import migrations, models

import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.userfield
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [("edc_appointment", "0005_auto_20161221_2135")]

    operations = [
        migrations.AlterField(
            model_name="appointment",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default="mac2-2.local",
                help_text="System field. (modified on create only)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="hostname_modified",
            field=edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="user_created",
            field=edc_model_fields.fields.userfield.UserField(
                blank=True, max_length=50, verbose_name="user created"
            ),
        ),
        migrations.AlterField(
            model_name="appointment",
            name="user_modified",
            field=edc_model_fields.fields.userfield.UserField(
                blank=True, max_length=50, verbose_name="user modified"
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="created",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default="mac2-2.local",
                help_text="System field. (modified on create only)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="hostname_modified",
            field=edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="modified",
            field=models.DateTimeField(blank=True, default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="user_created",
            field=edc_model_fields.fields.userfield.UserField(
                blank=True, max_length=50, verbose_name="user created"
            ),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="user_modified",
            field=edc_model_fields.fields.userfield.UserField(
                blank=True, max_length=50, verbose_name="user modified"
            ),
        ),
    ]



from django import forms
from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist

from edc_constants.constants import DWTA, OTHER
from edc_form_validators import FormValidator
from edc_prn.modelform_mixins import PrnFormValidatorMixin
from edc_utils import convert_php_dateformat

from .constants import TRANSFERRED


class SubjectTransferFormValidator(PrnFormValidatorMixin, FormValidator):
    """For use with the SubjectTransferForm"""

    def clean(self):
        # verify on study

        self.validate_other_specify("initiated_by", other_specify_field="initiated_by_other")

        self.m2m_single_selection_if(DWTA, m2m_field="transfer_reason")

        self.m2m_other_specify(
            OTHER, m2m_field="transfer_reason", field_other="transfer_reason_other"
        )


class SubjectTransferFormValidatorMixin:
    """Used in off schedule form or any form that
    needs to confirm the Subject Transfer form was submitted
    first.

    Note: This mixin is NOT for use with the SubjectTransferForm.
    """

    subject_transfer_model = None  # "inte_prn.subjecttransfer"
    subject_transfer_date_field = "transfer_date"
    subject_transfer_reason = TRANSFERRED

    @property
    def subject_transfer_model_cls(self):
        return django_apps.get_model(self.subject_transfer_model)

    def validate_subject_transferred(self):
        if self.subject_identifier:
            try:
                subject_transfer_obj = django_apps.get_model(
                    self.subject_transfer_model
                ).objects.get(subject_identifier=self.subject_identifier)
            except ObjectDoesNotExist as e:
                if (
                    self.cleaned_data.get(self.offschedule_reason_field)
                    and self.cleaned_data.get(self.offschedule_reason_field).name
                    == self.subject_transfer_reason
                ):
                    msg = (
                        "Patient has been transferred, please complete "
                        f"`{self.subject_transfer_model_cls._meta.verbose_name}` "
                        "form first."
                    )
                    raise forms.ValidationError({self.offschedule_reason_field: msg}) from e
            else:
                if self.cleaned_data.get(self.subject_transfer_date_field) and (
                    subject_transfer_obj.transfer_date
                    != self.cleaned_data.get(self.subject_transfer_date_field)
                ):
                    expected = subject_transfer_obj.transfer_date.strftime(
                        convert_php_dateformat(settings.SHORT_DATE_FORMAT)
                    )
                    got = self.cleaned_data.get(self.subject_transfer_date_field).strftime(
                        convert_php_dateformat(settings.SHORT_DATE_FORMAT)
                    )
                    raise forms.ValidationError(
                        {
                            self.subject_transfer_date_field: (
                                "Date does not match "
                                f"`{self.subject_transfer_model_cls._meta.verbose_name}` "
                                f"form. Expected {expected}. Got {got}."
                            )
                        }
                    )



from django.core.checks import CheckMessage, Error, Warning

from .consent_definition import ConsentDefinition
from .exceptions import ConsentDefinitionError
from .site_consents import site_consents


def check_consents(app_configs, **kwargs) -> list[CheckMessage]:
    errors = []
    errors.extend(check_consents_cdef_registered())
    errors.extend(check_consents_models())
    if not errors:
        errors.extend(check_consents_versions())
        errors.extend(check_consents_durations())
    return errors


def check_consents_cdef_registered() -> list[CheckMessage]:
    errors = []
    if not site_consents.registry:
        errors.append(
            Error("No consent definitions have been registered.", id="edc_consent.E001")
        )
    return errors


def check_consents_models() -> list[CheckMessage]:
    """Expect proxy models only in ConsentDefinitions.

    - ConsentDefinition may only be associated with a proxy model.
    - check proxy models use custom 'objects' and 'on_site' model
      managers.
    """
    errors = []
    for cdef in site_consents.registry.values():
        try:
            cdef.model_cls
        except ConsentDefinitionError as e:
            errors.append(Error(str(e), id="edc_consent.E002"))
        else:
            if not cdef.model_cls._meta.proxy:
                errors.append(
                    Error(
                        (
                            f"Consent definition model is not a proxy model. Got {cdef.model}."
                            f"See {cdef.name}"
                        ),
                        id="edc_consent.E003",
                    )
                )
    return errors


def check_consents_versions() -> list[CheckMessage]:
    """Expect versions to be unique across `proxy_for` model"""
    errors = []
    used = []
    for cdef in [cdef for cdef in site_consents.registry.values()]:
        if cdef in used:
            continue
        err, used = _inspect_others_using_same_proxy_for_model_with_duplicate_versions(
            cdef, used
        )
        if err:
            errors.append(err)
    return errors


def check_consents_durations() -> list[CheckMessage]:
    """Durations may not overlap across `proxy_for` model

    This check needs models to be ready otherwise we would add it
    to site_consents.register.
    """
    errors = []
    found = []
    cdefs: list[ConsentDefinition] = [cdef for cdef in site_consents.registry.values()]
    for cdef1 in cdefs:
        for cdef2 in cdefs:
            if cdef1 == cdef2:
                continue
            err, found = _inspect_possible_overlap_in_validity_period(cdef1, cdef2, found)
            if err:
                errors.append(err)
    return errors


def _inspect_possible_overlap_in_validity_period(
    cdef1, cdef2, found
) -> tuple[Warning | None, list]:
    """Durations between cdef1 and cdef2 may not overlap
    if they are using proxies of the same model -- `proxy_for` model.

    This is just a warning as there may be valid cases to allow this.
    For example, where consent definitions are customized by site.
    """
    err = None
    if (
        cdef1.model_cls._meta.proxy
        and cdef1.model_cls._meta.proxy_for_model == cdef1.model_cls._meta.proxy_for_model
    ):
        if cdef1.start <= cdef2.start <= cdef1.end or cdef1.start <= cdef2.end <= cdef1.end:
            if sorted([cdef1, cdef2], key=lambda x: x.version) in found:
                pass
            else:
                found.append(sorted([cdef1, cdef2], key=lambda x: x.version))
                err = Warning(
                    "Consent definition duration overlap found for same proxy_for_model. "
                    f"Got {cdef1.name} and {cdef2.name}.",
                    id="edc_consent.W002",
                )

    return err, found


def _inspect_others_using_same_proxy_for_model_with_duplicate_versions(
    cdef1: ConsentDefinition, used: list[ConsentDefinition]
) -> tuple[Warning | None, list[ConsentDefinition]]:
    err = None
    versions = []
    opts1 = cdef1.model_cls._meta
    for cdef2 in site_consents.registry.values():
        opts2 = cdef2.model_cls._meta
        if opts2.proxy and opts2.proxy_for_model == opts1.proxy_for_model:
            versions.append(cdef2.version)
            used.append(cdef2)
    if duplicates := [v for v in versions if versions.count(v) > 1]:
        err = Warning(
            f"Duplicate consent definition 'version' found for same proxy_for_model. "
            f"Got '{opts1.proxy_for_model._meta.label_lower}' versions {set(duplicates)}.",
            id="edc_consent.W001",
        )
    return err, used



from __future__ import annotations

import contextlib
from datetime import date, datetime
from zoneinfo import ZoneInfo

from dateutil.relativedelta import relativedelta
from django.conf import settings
from django.utils import timezone


class AgeValueError(Exception):
    pass


class AgeFormatError(Exception):
    pass


TWO_MONTHS = 2
TWELVE_MONTHS = 12


def get_dob(age_in_years: int, now: date | datetime | None = None) -> date:
    """Returns a DoB for the given age relative to now.

    Used in tests.
    """
    now = now or timezone.now()
    with contextlib.suppress(AttributeError):
        now = now.date()
    return now - relativedelta(years=age_in_years)


def age(born: date | datetime, reference_dt: date | datetime) -> relativedelta:
    """Returns a relative delta.

    Convert dates or datetimes to UTC datetimes.
    """
    if born is None:
        raise AgeValueError("DOB cannot be None")
    try:
        born_utc = born.astimezone(ZoneInfo("UTC"))
    except AttributeError:
        born_utc = datetime(*[*born.timetuple()][0:6], tzinfo=ZoneInfo("UTC"))
    try:
        reference_dt_utc = reference_dt.astimezone(ZoneInfo("UTC"))
    except AttributeError:
        reference_dt_utc = datetime(*[*reference_dt.timetuple()][0:6], tzinfo=ZoneInfo("UTC"))
    rdelta = relativedelta(reference_dt_utc, born_utc)
    if born_utc > reference_dt_utc:
        raise AgeValueError(
            f"Reference date {reference_dt} {reference_dt.tzinfo!s} "
            f"precedes DOB {born} {timezone}. Got {rdelta}"
        )
    return rdelta


def formatted_age(
    born: date | datetime | None,
    reference_dt: date | datetime,
    tz: str | None = None,
) -> str:
    age_as_str = "?"
    if born:
        tz = tz or getattr(settings, "TIME_ZONE", "UTC")
        born = datetime(*[*born.timetuple()][0:6], tzinfo=ZoneInfo(tz))
        reference_dt = reference_dt or timezone.now()
        age_delta = age(born, reference_dt or timezone.now())
        if age_delta.years == 0 and age_delta.months <= 0:
            age_as_str = f"{age_delta.days}d"
        elif age_delta.years == 0 and 0 < age_delta.months <= TWO_MONTHS:
            age_as_str = f"{age_delta.months}m{age_delta.days}d"
        elif age_delta.years == 0 and age_delta.months > TWO_MONTHS:
            age_as_str = f"{age_delta.months}m"
        elif age_delta.years == 1:
            m = age_delta.months + TWELVE_MONTHS
            age_as_str = f"{m}m"
        else:
            age_as_str = f"{age_delta.years}y"
    return age_as_str


def get_age_in_days(reference_datetime: date | datetime, dob: date | datetime) -> int:
    age_delta = age(dob, reference_datetime)
    return age_delta.days



from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from django_audit_fields import audit_fieldset_tuple

from edc_model_admin.history import SimpleHistoryAdmin
from edc_pharmacy.admin.model_admin_mixin import ModelAdminMixin
from edc_pharmacy.admin.remove_fields_for_blinded_users import (
    remove_fields_for_blinded_users,
)
from edc_pharmacy.admin_site import edc_pharmacy_admin
from edc_pharmacy.forms import ConfirmationForm
from edc_pharmacy.models import Confirmation
from edc_utils.date import to_local


@admin.register(Confirmation, site=edc_pharmacy_admin)
class ConfirmationAdmin(ModelAdminMixin, SimpleHistoryAdmin):

    change_list_title = "Pharmacy: Confirmed stock"
    change_form_title = "Pharmacy: Confirmed stock"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    form = ConfirmationForm

    actions = ["delete_selected"]

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "confirmation_identifier",
                    "confirmed_datetime",
                    "confirmed_by",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "identifier",
        "stock_changelist",
        "confirmed_date",
        "confirmed_by",
    )

    readonly_fields = (
        "confirmed_datetime",
        "confirmed_by",
    )

    def get_list_display(self, request):
        fields = super().get_list_display(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields

    def get_list_filter(self, request):
        fields = super().get_list_filter(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields

    def get_search_fields(self, request):
        fields = super().get_search_fields(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields

    @admin.display(description="CONFIRMED #", ordering="confirmation_identifier")
    def identifier(self, obj):
        return obj.confirmation_identifier.split("-")[0]

    @admin.display(description="Confirmed date", ordering="confirmed_datetime")
    def confirmed_date(self, obj):
        return to_local(obj.confirmed_datetime).date()

    @admin.display(description="Stock #")
    def stock_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
        url = f"{url}?q={obj.stock.code}"
        context = dict(url=url, label=f"{obj.stock.code}", title="Go to stock")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist

from ..utils import get_rxrefill_model_cls

if TYPE_CHECKING:
    from ..models import Rx, RxRefill


def get_active_refill(rx: Rx) -> RxRefill | None:
    """Returns the 'active' Refill instance or None
    for this prescription.
    """
    try:
        rx_refill = get_rxrefill_model_cls().objects.get(rx=rx, active=True)
    except ObjectDoesNotExist:
        rx_refill = None
    except MultipleObjectsReturned:
        rx_refill = (
            get_rxrefill_model_cls()
            .objects.filter(rx=rx, active=True)
            .order_by("refill_start_datetime")
            .last()
        )
        get_rxrefill_model_cls().objects.filter(rx=rx, active=True).exclude(
            pk=rx_refill.pk
        ).update(active=False)
    return rx_refill



from django.core.validators import RegexValidator
from django.db import models


class AliquotTypeModelMixin(models.Model):
    aliquot_type = models.CharField(verbose_name="Aliquot Type Name", max_length=25)

    alpha_code = models.CharField(
        verbose_name="Aliquot Type Alpha Code",
        validators=[RegexValidator("^[A-Z]{2}$")],
        max_length=25,
    )

    numeric_code = models.CharField(
        verbose_name="Aliquot Type Numeric Code",
        validators=[RegexValidator("^[0-9]{2}$")],
        max_length=25,
    )

    class Meta:
        abstract = True



from __future__ import annotations

import copy
import sys
from collections.abc import Callable
from importlib import import_module
from typing import Any

from django.apps import apps as django_apps
from django.core.management.color import color_style
from django.utils.module_loading import module_has_submodule


class AlreadyRegistered(Exception):
    pass


class SitePharmacyError(Exception):
    pass


class LabelConfig:
    def __init__(
        self,
        name: str,
        drawing_callable: Callable,
        label_cls: Any,
        test_data_func: Callable | None = None,
    ):
        self.name = name
        self.drawing_callable = drawing_callable
        self.label_cls = label_cls
        self.test_data_func = test_data_func or self._test_data_func

    def _test_data_func(self) -> dict:
        return {}

    def __repr__(self):
        return self.__class__.__name__ + f"({self.name})"

    def __str__(self):
        return self.__class__.__name__ + f"({self.name})"


class SiteLabelConfigs:
    def __init__(self):
        self.registry: dict[str, LabelConfig] = {}

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __len__(self):
        return len(self.registry.values())

    def __iter__(self):
        return iter(self.registry.values())

    def unregister(self, key) -> None:
        if key in self.registry:
            del self.registry[key]

    def register(
        self,
        name: str,
        drawing_callable: Callable,
        label_cls: Any,
        test_data_func: Callable | None = None,
    ) -> None:
        if name in self.registry:
            raise AlreadyRegistered(f"Already registered. Got name='{name}' ")
        self.registry.update(
            {name: LabelConfig(name, drawing_callable, label_cls, test_data_func)}
        )

    def all(self) -> dict[str, LabelConfig]:
        return self.registry

    def get(self, name) -> LabelConfig:
        if name not in self.registry:
            raise SitePharmacyError(
                f"Name does not exist. Is it registered? "
                f"Expected one of {self.registry}. Got {name}."
            )
        return self.registry.get(name)

    @staticmethod
    def autodiscover(module_name=None, verbose=True):
        before_import_registry = {}
        module_name = module_name or "label_configs"
        writer = sys.stdout.write if verbose else lambda x: x
        style = color_style()
        writer(f" * checking for site {module_name} ...\n")
        for app in django_apps.app_configs:
            writer(f" * searching {app}           \r")
            try:
                mod = import_module(app)
                try:
                    before_import_registry = copy.copy(site_label_configs.registry)
                    import_module(f"{app}.{module_name}")
                    writer(f"   - registered '{module_name}' from '{app}'\n")
                except SitePharmacyError as e:
                    writer(f"   - loading {app}.{module_name} ... ")
                    writer(style.ERROR(f"ERROR! {e}\n"))
                except ImportError as e:
                    site_label_configs.registry = before_import_registry
                    if module_has_submodule(mod, module_name):
                        raise SitePharmacyError(str(e))
            except ImportError:
                pass


site_label_configs = SiteLabelConfigs()



from zoneinfo import ZoneInfo

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils import timezone

from edc_identifier.model_mixins import UniqueSubjectIdentifierFieldMixin
from edc_model.validators import datetime_not_future
from edc_model_fields.fields import OtherCharField
from edc_protocol.validators import datetime_not_before_study_start
from edc_utils import convert_php_dateformat
from edc_visit_schedule.utils import (
    off_all_schedules_or_raise,
    offstudy_datetime_after_all_offschedule_datetimes,
)

from ..choices import OFFSTUDY_REASONS
from ..utils import OffstudyError


class OffstudyModelMixinError(ValidationError):
    pass


class OffstudyModelMixin(UniqueSubjectIdentifierFieldMixin, models.Model):
    """Model mixin for the Off-study model.

    Override in admin like this:

        def formfield_for_choice_field(self, db_field, request, **kwargs):
            if db_field.name == "offstudy_reason":
                kwargs['choices'] = OFFSTUDY_REASONS
            return super().formfield_for_choice_field(db_field, request, **kwargs)

    """

    offstudy_reason_choices = OFFSTUDY_REASONS

    offstudy_datetime = models.DateTimeField(
        verbose_name="Off-study date and time",
        validators=[datetime_not_before_study_start, datetime_not_future],
        default=timezone.now,
    )

    report_datetime = models.DateTimeField(null=True, editable=False)

    offstudy_reason = models.CharField(
        verbose_name="Please code the primary reason participant taken off-study",
        choices=offstudy_reason_choices,
        max_length=125,
    )

    other_offstudy_reason = OtherCharField()

    comment = models.TextField(
        verbose_name="Please provide further details if possible",
        max_length=500,
        blank=True,
        default="",
    )

    def __str__(self):
        dte_str = self.report_datetime.astimezone(ZoneInfo(settings.TIME_ZONE)).strftime(
            convert_php_dateformat(settings.SHORT_DATETIME_FORMAT)
        )
        return f"{self.subject_identifier} {dte_str}"

    def save(self, *args, **kwargs):
        self.report_datetime = self.offstudy_datetime
        datetime_not_before_study_start(self.offstudy_datetime)
        datetime_not_future(self.offstudy_datetime)
        off_all_schedules_or_raise(subject_identifier=self.subject_identifier)
        offstudy_datetime_after_all_offschedule_datetimes(
            subject_identifier=self.subject_identifier,
            offstudy_datetime=self.offstudy_datetime,
            exception_cls=OffstudyError,
        )
        super().save(*args, **kwargs)

    def natural_key(self):
        return (self.subject_identifier,)

    class Meta:
        abstract = True



from urllib.parse import urlparse

from django.core.mail.message import EmailMessage
from mempass.password_generator import PasswordGenerator

from edc_dashboard.utils import get_index_page
from edc_protocol.research_protocol_config import ResearchProtocolConfig

from .import_users import change_user_template


def change_password(user, nwords: int | None = None) -> str:
    nwords = nwords or 4
    pwdgen = PasswordGenerator(nwords=nwords)
    password = pwdgen.get_password()
    user.set_password(password)
    user.save()
    return password


def send_new_credentials_to_user(user, nwords: int | None = None) -> EmailMessage:
    body = change_user_template
    site_names = "\n - ".join([s.name for s in user.userprofile.sites.all()])
    role_names = "\n - ".join([r.display_name for r in user.userprofile.roles.all()])
    resource_name = urlparse(get_index_page()).netloc
    project_name = ResearchProtocolConfig().project_name
    password = change_password(user, nwords)
    opts = {
        "resource_name": resource_name,
        "project_name": project_name,
        "username": user.username,
        "password": password,
        "first_name": user.first_name,
        "last_name": user.last_name,
        "job_title": user.userprofile.job_title or "?job title?",
        "email": user.email,
        "site_names": site_names or "None",
        "role_names": role_names or "None",
    }
    return EmailMessage(
        f"{project_name} EDC: Your {resource_name} user account has been updated.",
        body=body.safe_substitute(opts),
        from_email="noreply@clinicedc.org",
        to=(user.email,),
    ).send(fail_silently=True)



from edc_constants.constants import (
    CLOSED,
    FEEDBACK,
    HIGH_PRIORITY,
    NEW,
    NORMAL,
    OPEN,
    RESOLVED,
)

from .constants import CLOSED_WITH_ACTION

RESPONSE_STATUS = (
    (NEW, "New"),
    (OPEN, "Open"),
    (FEEDBACK, "Feedback, awaiting data manager"),
    (RESOLVED, "Resolved"),
)


DM_STATUS = (
    (OPEN, "Open, awaiting site"),
    (CLOSED, "Closed"),
    (CLOSED_WITH_ACTION, "Closed, with plan of action"),
)

QUERY_PRIORITY = ((HIGH_PRIORITY, "High"), (NORMAL, "Normal"))



from .constants import KEYED, MISSED, NOT_REQUIRED, REQUIRED

ENTRY_CATEGORY = (("CLINIC", "Clinic"), ("LAB", "Lab"), ("OTHER", "Other"))

ENTRY_STATUS = (
    (REQUIRED, "New"),
    (KEYED, "Keyed"),
    (MISSED, "Missed"),
    (NOT_REQUIRED, "Not required"),
)

ENTRY_WINDOW = (("VISIT", "Visit"), ("FORM", "Form"))

VISIT_INTERVAL_UNITS = (("H", "Hour"), ("D", "Day"), ("M", "Month"), ("Y", "Year"))

TAG_TYPE = (("REGISTRATION", "Registration"), ("OTHER", "Other"))



from django.db import models
from django_audit_fields.models import AuditUuidModelMixin

from .url_model_mixin import UrlModelMixin

default_permissions = ("add", "change", "delete", "view", "export", "import")


class BaseUuidModel(UrlModelMixin, AuditUuidModelMixin, models.Model):
    """Default manager must be `objects` for modelform validation to
    work correctly with unique fields.
    """

    objects = models.Manager()

    class Meta(AuditUuidModelMixin.Meta):
        abstract = True
        default_permissions = default_permissions
        default_manager_name = "objects"
        indexes = AuditUuidModelMixin.Meta.indexes



# Generated by Django 2.1 on 2018-10-18 23:07

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("edc_notification", "0002_update_notifications")]

    # this is now run as in post migrate
    # operations = [migrations.RunPython(site_notifications.update_notification_list)]

    operations = []



from django.contrib import admin
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from edc_model_admin.mixins import TemplatesModelAdminMixin

from ..admin_site import edc_auth_admin
from ..forms import UserProfileForm
from ..models import UserProfile
from .fieldsets import user_profile_fieldsets


class UserProfileInline(admin.StackedInline):
    model = UserProfile
    can_delete = False
    verbose_name_plural = "User profile"

    form = UserProfileForm

    fieldsets = user_profile_fieldsets

    filter_horizontal = ("email_notifications", "sms_notifications", "sites", "roles")


@admin.register(UserProfile, site=edc_auth_admin)
class UserProfileAdmin(TemplatesModelAdminMixin, admin.ModelAdmin):
    fieldsets = user_profile_fieldsets

    filter_horizontal = ("email_notifications", "sms_notifications", "sites")

    list_display = (
        "user",
        "user_sites",
        "user_email_notifications",
        "user_sms_notifications",
        "mobile",
        "export_format",
        "is_multisite_viewer",
    )

    list_filter = ("sites__name", "email_notifications__name", "is_multisite_viewer")

    search_fields = ("user__username", "mobile", "user__email")

    @staticmethod
    def user_sites(obj=None):
        site_names = [o.name for o in obj.sites.all().order_by("name")]
        return format_html("{}", mark_safe("<BR>".join(site_names)))  # nosec B703 B308

    @staticmethod
    def user_email_notifications(obj=None):
        display_names = [
            o.display_name for o in obj.email_notifications.all().order_by("display_name")
        ]
        return format_html("{}", mark_safe("<BR>".join(display_names)))  # nosec B703 B308

    @staticmethod
    def user_sms_notifications(obj=None):
        display_names = [
            o.display_name for o in obj.sms_notifications.all().order_by("display_name")
        ]
        return format_html(
            "{}",
            mark_safe("<BR>".join(display_names)),  # nosec B703, B308
        )



# Generated by Django 2.2.2 on 2019-09-22 02:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_visit_schedule", "0006_auto_20190706_0706")]

    operations = [
        migrations.AlterField(
            model_name="subjectschedulehistory",
            name="subject_identifier",
            field=models.CharField(max_length=50),
        )
    ]



from django.db import models

from edc_reportable import MILLIMOLES_PER_LITER_DISPLAY
from edc_reportable.units import MILLIGRAMS_PER_DECILITER, MILLIMOLES_PER_LITER

from ..model_mixin_factories import reportable_result_model_mixin_factory


class MagnesiumModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="magnesium",
        verbose_name="Magnesium",
        units_choices=(
            (MILLIMOLES_PER_LITER, MILLIMOLES_PER_LITER_DISPLAY),
            (MILLIGRAMS_PER_DECILITER, MILLIGRAMS_PER_DECILITER),
        ),
        decimal_places=2,
    ),
    models.Model,
):
    class Meta:
        abstract = True


class PotassiumModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="potassium",
        verbose_name="Potassium",
        units_choices=((MILLIMOLES_PER_LITER, MILLIMOLES_PER_LITER_DISPLAY),),
        decimal_places=1,
    ),
    models.Model,
):
    class Meta:
        abstract = True


class SodiumModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="sodium",
        verbose_name="Sodium (Na)",
        units_choices=((MILLIMOLES_PER_LITER, MILLIMOLES_PER_LITER_DISPLAY),),
        decimal_places=0,
    ),
    models.Model,
):
    class Meta:
        abstract = True



from django.contrib import admin, messages
from django.db.models import QuerySet
from django.utils.translation import gettext

from ...models import OrderItem, Stock


@admin.display(description=f"Delete selected {OrderItem._meta.verbose_name_plural}")
def delete_order_items_action(modeladmin, request, queryset: QuerySet[OrderItem]):
    failed_count = 0
    success_count = 0
    for obj in queryset:
        if obj.receiveitem_set.filter(stock__confirmation__isnull=False).exists():
            failed_count += 1
        else:
            Stock.objects.filter(
                receive_item__order_item=obj, stock__confirmation__isnull=True
            ).delete()
            obj.receiveitem_set.all().delete()
            obj.delete()
            success_count += 1
    if success_count > 0:
        messages.add_message(
            request,
            messages.SUCCESS,
            gettext(f"Successfully deleted {success_count} {OrderItem._meta.verbose_name}."),
        )
    if failed_count > 0:
        messages.add_message(
            request,
            messages.ERROR,
            gettext(
                f"Unable to deleted {failed_count} {OrderItem._meta.verbose_name}. "
                "Confirmed stock items exist."
            ),
        )



BLOOD_GLUCOSE = "blood_glucose"
BLOOD_GLUCOSE_POC = "blood_glucose_poc"
CD4 = "CD4"
CHEMISTRY_LFT = "chemistry_lft"
CHEMISTRY_LIPIDS = "chemistry_lipids"
CHEMISTRY_RFT = "chemistry_rft"
CHOL = "chol"
FBC = "fbc"
HAEMOGLOBIN = "haemoglobin"
HBA1C = "hba1c"
HBA1C_POC = "hba1c_poc"
HCT = "hct"
HDL = "hdl"
INSULIN = "insulin"
LDL = "ldl"
LFT = "LFT"
LIPIDS = "LIPIDS"
LYMPHOCYTE = "lymphocyte"
LYMPHOCYTE_DIFF = "lymphocyte_diff"
MCH = "mch"
MCHC = "mchc"
MCV = "mcv"
NEUTROPHIL = "neutrophil"
NEUTROPHIL_DIFF = "neutrophil_diff"
PLATELETS = "platelets"
RBC = "rbc"
RFT = "RFT"
SPUTUM = "sputum"
TRIG = "trig"
VL = "viral_load"
WBC = "wbc"



# Generated by Django 5.2.6 on 2025-09-22 12:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_lab", "0037_alter_historicalorder_order_datetime_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="aliquot",
            name="slug",
            field=models.CharField(
                db_index=True,
                default="",
                editable=False,
                help_text="Hold slug field values for quick search. Excludes encrypted fields",
                max_length=250,
            ),
        ),
        migrations.AlterField(
            model_name="box",
            name="slug",
            field=models.CharField(
                db_index=True,
                default="",
                editable=False,
                help_text="Hold slug field values for quick search. Excludes encrypted fields",
                max_length=250,
            ),
        ),
        migrations.AlterField(
            model_name="boxitem",
            name="slug",
            field=models.CharField(
                db_index=True,
                default="",
                editable=False,
                help_text="Hold slug field values for quick search. Excludes encrypted fields",
                max_length=250,
            ),
        ),
        migrations.AlterField(
            model_name="historicalaliquot",
            name="slug",
            field=models.CharField(
                db_index=True,
                default="",
                editable=False,
                help_text="Hold slug field values for quick search. Excludes encrypted fields",
                max_length=250,
            ),
        ),
        migrations.AlterField(
            model_name="historicalbox",
            name="slug",
            field=models.CharField(
                db_index=True,
                default="",
                editable=False,
                help_text="Hold slug field values for quick search. Excludes encrypted fields",
                max_length=250,
            ),
        ),
        migrations.AlterField(
            model_name="historicalboxitem",
            name="slug",
            field=models.CharField(
                db_index=True,
                default="",
                editable=False,
                help_text="Hold slug field values for quick search. Excludes encrypted fields",
                max_length=250,
            ),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="slug",
            field=models.CharField(
                db_index=True,
                default="",
                editable=False,
                help_text="Hold slug field values for quick search. Excludes encrypted fields",
                max_length=250,
            ),
        ),
        migrations.AlterField(
            model_name="manifest",
            name="slug",
            field=models.CharField(
                db_index=True,
                default="",
                editable=False,
                help_text="Hold slug field values for quick search. Excludes encrypted fields",
                max_length=250,
            ),
        ),
        migrations.AlterField(
            model_name="manifestitem",
            name="slug",
            field=models.CharField(
                db_index=True,
                default="",
                editable=False,
                help_text="Hold slug field values for quick search. Excludes encrypted fields",
                max_length=250,
            ),
        ),
    ]



import re

from django.apps import apps as django_apps

from edc_constants.constants import UUID_PATTERN
from edc_lab.lab import Specimen as SpecimenObject
from edc_lab.site_labs import site_labs
from edc_label import LabPrintersMixin, add_job_results_to_messages


class ProcessRequisitionViewMixin(LabPrintersMixin):
    def process(self, request=None):
        """Creates aliquots according to the lab_profile.

        Actions handled by the Specimen object.
        """
        processed = {}
        for requisition in self.get_requisitions(
            pk__in=self.selected_items, received=True, processed=False
        ):
            specimen = SpecimenObject(requisition=requisition)
            if requisition.panel_object.processing_profile:
                processed.update({"requisition": specimen.process()})
                requisition.processed = True
                requisition.save()
        for created_aliquots in processed.values():
            pks = [specimen.primary_aliquot.pk] + [obj.pk for obj in created_aliquots]
            if pks:
                job_result = self.print_labels(pks=pks, request=request)
                if job_result:
                    add_job_results_to_messages(request, [job_result])

    @property
    def selected_items(self):
        """Returns a list of UUIDs as strings."""
        if not self._selected_items:
            for pk in self.request.POST.getlist(self.form_action_selected_items_name):
                if re.match(UUID_PATTERN, str(pk)):
                    self._selected_items.append(pk)
        return self._selected_items

    def get_requisitions(self, **kwargs):
        """Returns a list of requisition model instances filtered
        by the given kwargs.

        Searches all requisition model classes registered with
        site_labs.
        """
        requisitions = []
        for model in site_labs.requisition_models.values():
            model_cls = django_apps.get_model(model)
            requisitions.extend([r for r in model_cls.objects.filter(**kwargs)])
        return requisitions

    @property
    def processed_requisitions(self):
        """Returns a list of model instances that have been
        processed.
        """
        return self.get_requisitions(pk__in=self.selected_items, processed=True)

    @property
    def unprocessed_requisitions(self):
        """Returns a list of model instances that have NOT been
        processed.
        """
        return self.get_requisitions(pk__in=self.selected_items, processed=False)



from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import CharField, TextField

from edc_constants.constants import NULL_STRING
from edc_model import DEFAULT_BASE_FIELDS

from ..utils import get_registered_subject_model_cls

if TYPE_CHECKING:
    from edc_registration.models import RegisteredSubject


class UpdatesOrCreatesRegistrationModelError(Exception):
    pass


class UpdatesOrCreatesRegistrationModelMixin(models.Model):
    """A model mixin that creates or updates Registration model
    (e.g. RegisteredSubject) on post_save signal.
    """

    @property
    def registration_model(self) -> type[RegisteredSubject]:
        """Returns the Registration model"""
        return get_registered_subject_model_cls()

    def registration_update_or_create(self) -> tuple[RegisteredSubject, bool]:
        """Creates or Updates the registration model with attributes
        from this instance.

        Called from the signal.

        Note: `registration_unique_field` is typically
              "subject_identifier".
        """
        if not getattr(self, self.registration_unique_field):
            raise UpdatesOrCreatesRegistrationModelError(
                f"Cannot update or create Registration model. "
                f"Field value for '{self.registration_unique_field}' is None."
            )

        registration_value = getattr(self, self.registration_unique_field)
        registration_value = self.to_string(registration_value)
        try:
            obj = self.registration_model.objects.get(
                **{self.registered_model_unique_field: registration_value}
            )
        except ObjectDoesNotExist:
            pass
        else:
            self.registration_raise_on_illegal_value_change(obj)
        registration_obj, created = self.registration_model.objects.update_or_create(
            **{self.registered_model_unique_field: registration_value},
            defaults=self.registration_options,
        )
        return registration_obj, created

    @staticmethod
    def to_string(value) -> str:
        """Returns a string.

        Converts UUID to string using .hex.
        """
        with contextlib.suppress(AttributeError):
            value = str(value.hex)
        return value

    @property
    def registration_unique_field(self) -> str:
        """Returns the field attr on YOUR model that will update
        `registered_model_unique_field`.

        Typically, `subject_identifier`.
        """
        return "subject_identifier"

    @property
    def registered_model_unique_field(self) -> str:
        """Returns the field attr on THIS model to be queried against
        the value of `registration_unique_field`.
        """
        return self.registration_unique_field

    def registration_raise_on_illegal_value_change(self, registration_obj):
        """Raises an exception if a value changes between updates.

        Values are available in `registration_options`.
        """
        pass

    @property
    def registration_options(self) -> dict:
        """Gathers values for common attributes between the
        registration model and this instance.
        """
        registration_options = {}
        rs = self.registration_model()
        for fname, value in self.__dict__.items():
            if fname not in (*DEFAULT_BASE_FIELDS, "_state"):
                try:
                    getattr(rs, fname)
                except AttributeError:
                    pass
                else:
                    value = (  # noqa: PLW2901
                        NULL_STRING
                        if value is None
                        and isinstance(rs._meta.get_field(fname), (CharField, TextField))
                        else value
                    )
                    registration_options.update({fname: value})
        registration_identifier = registration_options.get("registration_identifier")
        if registration_identifier:
            registration_options["registration_identifier"] = self.to_string(
                registration_identifier
            )
        return registration_options

    class Meta:
        abstract = True



# Generated by Django 3.2.4 on 2021-07-14 21:55

import _socket
import django.db.models.deletion
import django_audit_fields.fields.hostname_modification_field
import django_audit_fields.fields.userfield
import django_audit_fields.fields.uuid_auto_field
import django_audit_fields.models.audit_model_mixin
import django_revision.revision_field
from django.db import migrations, models

import edc_model.models.fields.other_charfield
import edc_sites.models
import django.utils.timezone


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("sites", "0002_alter_domain_unique"),
    ]

    operations = [
        migrations.CreateModel(
            name="RefusalReasons",
            fields=[
                (
                    "name",
                    models.CharField(
                        db_index=True,
                        help_text="This is the stored value, required",
                        max_length=250,
                        unique=True,
                        verbose_name="Stored value",
                    ),
                ),
                (
                    "display_name",
                    models.CharField(
                        db_index=True,
                        help_text="(suggest 40 characters max.)",
                        max_length=250,
                        unique=True,
                        verbose_name="Name",
                    ),
                ),
                (
                    "display_index",
                    models.IntegerField(
                        db_index=True,
                        default=0,
                        help_text="Index to control display order if not alphabetical, not required",
                        verbose_name="display index",
                    ),
                ),
                (
                    "field_name",
                    models.CharField(
                        blank=True,
                        editable=False,
                        help_text="Not required",
                        max_length=25,
                        null=True,
                    ),
                ),
                (
                    "version",
                    models.CharField(default="1.0", editable=False, max_length=35),
                ),
                ("id", models.AutoField(primary_key=True, serialize=False)),
            ],
            options={
                "verbose_name": "Refusal Reason",
                "verbose_name_plural": "Refusal Reasons",
                "ordering": ["display_index", "display_name"],
                "abstract": False,
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
            },
        ),
        migrations.CreateModel(
            name="SubjectRefusal",
            fields=[
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "modified",
                    models.DateTimeField(
                        blank=True,
                        default=django.utils.timezone.now,
                    ),
                ),
                (
                    "user_created",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    django_audit_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("screening_identifier", models.CharField(max_length=50, unique=True)),
                (
                    "report_datetime",
                    models.DateTimeField(
                        default=django.utils.timezone.now,
                        verbose_name="Report Date and Time",
                    ),
                ),
                (
                    "other_reason",
                    edc_model.models.fields.other_charfield.OtherCharField(
                        blank=True,
                        max_length=35,
                        null=True,
                        verbose_name="If other, please specify ...",
                    ),
                ),
                (
                    "comment",
                    models.TextField(
                        blank=True, null=True, verbose_name="Additional Comments"
                    ),
                ),
                (
                    "reason",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="edc_refusal.refusalreasons",
                        verbose_name="Reason for refusal to join",
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="sites.site",
                    ),
                ),
            ],
            options={
                "verbose_name": "Subject Refusal",
                "verbose_name_plural": "Subject Refusals",
                "ordering": ("-modified", "-created"),
                "get_latest_by": "modified",
                "abstract": False,
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
            },
            managers=[
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
        migrations.AddIndex(
            model_name="refusalreasons",
            index=models.Index(
                fields=["id", "display_name", "display_index"],
                name="edc_refusal_id_be47bd_idx",
            ),
        ),
    ]



from django import forms

from edc_constants.constants import NO, YES

___all__ = ["ReviewFieldsModelFormMixin"]


class ReviewFieldsModelFormMixin:
    def clean_consent_reviewed(self) -> str:
        consent_reviewed = self.cleaned_data.get("consent_reviewed")
        if consent_reviewed != YES:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return consent_reviewed

    def clean_study_questions(self) -> str:
        study_questions = self.cleaned_data.get("study_questions")
        if study_questions != YES:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return study_questions

    def clean_assessment_score(self) -> str:
        assessment_score = self.cleaned_data.get("assessment_score")
        if assessment_score != YES:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return assessment_score

    def clean_consent_copy(self) -> str:
        consent_copy = self.cleaned_data.get("consent_copy")
        if consent_copy == NO:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return consent_copy

    def clean_consent_signature(self) -> str:
        consent_signature = self.cleaned_data.get("consent_signature")
        if consent_signature != YES:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return consent_signature



import json
import sys

from django.core.checks import Error
from django.db import OperationalError

from edc_sites.single_site import SingleSite
from edc_sites.site import SitesCheckError
from edc_sites.site import sites as site_sites
from edc_sites.utils import get_site_model_cls


def sites_check(app_configs, **kwargs):
    errors = []
    if "migrate" not in sys.argv and "makemigrations" not in sys.argv:
        try:
            compare_single_sites_with_db()
        except (SitesCheckError, OperationalError) as e:
            errors.append(
                Error(
                    e,
                    hint="Sites model is out-of-sync with edc_sites registry.",
                    obj=site_sites,
                    id="edc_sites.E001",
                )
            )
        if not site_sites.all():
            errors.append(
                Error(
                    "No sites have been registered",
                    id="edc_sites.E002",
                )
            )

    return errors


def compare_single_sites_with_db():
    """Checks the Site / SiteProfile tables are in sync"""
    if not get_site_model_cls().objects.all().exists():
        raise SitesCheckError("No sites have been imported. You need to run migrate")
    ids1 = sorted(list(site_sites.all()))
    ids2 = [x[0] for x in get_site_model_cls().objects.values_list("id").all().order_by("id")]
    if ids1 != ids2:
        raise SitesCheckError(
            f"Site table is out of sync. Got registered sites = {ids1}. "
            f"Sites in Sites model = {ids2}. Try running migrate."
        )
    for site_id, single_site in site_sites.all().items():
        site_obj = get_site_model_cls().objects.get(id=site_id)
        match_name_and_domain_or_raise(single_site, site_obj)
        match_country_and_country_code_or_raise(single_site, site_obj)
        match_languages_or_raise(single_site, site_obj)
        match_title_with_description_or_raise(single_site, site_obj)


def match_name_and_domain_or_raise(single_site: SingleSite, site_obj):
    for attr in ["name", "domain"]:
        value1 = getattr(single_site, attr)
        value2 = getattr(site_obj, attr)
        if value1 != value2:
            raise SitesCheckError(
                f"Site table is out of sync. Comparing {attr} of site `{site_obj.id}`. "
                "between the SingleSite and Site model. "
                f"Got `{value1}` != `{value2}`. Try running migrate."
            )


def match_country_and_country_code_or_raise(single_site: SingleSite, site_obj):
    for attr in ["country", "country_code"]:
        value1 = getattr(single_site, attr)
        value2 = getattr(site_obj.siteprofile, attr)
        if value1 != value2:
            raise SitesCheckError(
                f"Site table is out of sync. Checking {site_obj.id} {attr}. "
                f"Try running migrate. Got {value1} != {value2}"
            )


def match_languages_or_raise(single_site: SingleSite, site_obj):
    value1 = single_site.languages
    value2 = json.loads(site_obj.siteprofile.languages)
    if value1 != value2:
        raise SitesCheckError(
            f"Site table is out of sync. Checking {site_obj.id} "
            f"languages. Try running migrate. Got {value1} != {value2}"
        )


def match_title_with_description_or_raise(single_site: SingleSite, site_obj):
    value1 = site_obj.siteprofile.title
    value2 = single_site.description
    if value1 != value2:
        raise SitesCheckError(
            f"Site table is out of sync. Checking {site_obj.id} title/description. "
            f"Try running migrate. Got {value1} != {value2}"
        )



from django.contrib import admin
from django.contrib.sites.models import Site

from edc_model_admin.mixins import TemplatesModelAdminMixin

from ..admin_site import edc_sites_admin

__all__ = ["SiteAdmin"]

admin.site.unregister(Site)


@admin.register(Site, site=edc_sites_admin)
class SiteAdmin(TemplatesModelAdminMixin, admin.ModelAdmin):
    pass



from .clinical_review_followup_chol_model_mixin import ClinicalReviewCholModelMixin
from .clinical_review_followup_dm_model_mixin import ClinicalReviewDmModelMixin
from .clinical_review_followup_hiv_model_mixin import ClinicalReviewHivModelMixin
from .clinical_review_followup_htn_model_mixin import ClinicalReviewHtnModelMixin
from .clinical_review_followup_model_mixin import ClinicalReviewModelMixin



from django.conf import settings

from edc_lab.constants import SHIPPED

from .dashboard_templates import dashboard_templates
from .dashboard_urls import dashboard_urls


class DashboardMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        return response

    def process_view(self, request, *args):
        request.url_name_data.update(**dashboard_urls)
        try:
            url_name_data = settings.LAB_DASHBOARD_URL_NAMES
        except AttributeError:
            pass
        else:
            request.url_name_data.update(**url_name_data)

        template_data = dashboard_templates
        try:
            template_data.update(settings.LAB_DASHBOARD_BASE_TEMPLATES)
        except AttributeError:
            pass
        request.template_data.update(**template_data)

    def process_template_response(self, request, response):
        if response.context_data:
            response.context_data.update(**request.url_name_data)
            response.context_data.update(**request.template_data)
            response.context_data.update(SHIPPED=SHIPPED)
        return response



class ColumnApply:
    column_names = None

    def __init__(self, dataframe=None):
        self.dataframe = dataframe
        for column_name in self.column_names:
            if column_name in list(self.dataframe.columns):
                self._apply(column_name)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.column_names})"

    def _apply(self, column_name):
        self.dataframe[column_name] = self.dataframe[column_name].apply(self.apply)

    def apply(self, value):
        return value



from __future__ import annotations

from functools import partial

from django import forms
from django.contrib import messages
from django.contrib.admin.utils import flatten_fieldsets
from django.contrib.messages import MessageFailure
from django.core.exceptions import FieldError
from django.forms.models import modelform_defines_fields, modelform_factory
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from edc_sites.modelform_mixins import SiteModelFormMixin


class ModelAdminBypassDefaultFormClsMixin:
    """A modeladmin mixin to set the admin class's modelform conditionally
    based on a special data manager permission codename.

    Position this mixin just before `admin.ModelAdmin` to ensure the
    correct MRO for `get_form`.

    The custom modelform factory returns a basic modelform for the
    given admin model WITHOUT ANY validation checks.

    This special case can be used briefly, carefully and by an
    experienced user to update a form that cannot be updated
    because of a blocking validation check. The user MUST know
    what they are doing to avoid creating data integrity issues.

    For example:
        A screening form's validation check blocks the screening form
        from further edits because the subject's consent form has
        been submitted. The permission-based custom form feature can
        be used to bypass ALL validation on the screening form and
        allow the user to update the screening form.
    """

    # if custom_form_codename is None, this mixin does nothing
    custom_form_codename: str | None = None  # "e.g. edc_data_manager.special_bypassmodelform"

    def custom_modelform_factory(self):
        """Returns a basic modelform (with no validation checks)"""

        meta_cls = getattr(
            self.form,
            "Meta",
            type("Meta", (object,), {"model": self.model, "fields": "__all__"}),
        )

        class BasicModelForm(SiteModelFormMixin, forms.ModelForm):
            Meta = meta_cls

        BasicModelForm.__name__ = meta_cls.model.__name__ + "BasicForm"
        return BasicModelForm

    def get_form_cls(self, request):
        """Returns a custom Form class or the default.

        A custom Form class is returned if the user has the correct
        permissions otherwise returns the Form class originally
        declared on this ModelAdmin class.
        """
        if self.custom_form_codename and request.user.has_perm(self.custom_form_codename):
            try:
                messages.add_message(
                    request,
                    messages.WARNING,
                    format_html(
                        "{html}",
                        html=mark_safe(
                            "<B>WARNING: All validation checks for this form have been "
                            "disabled. </B><BR>Your user account includes the special "
                            f"permission `{self.custom_form_codename}`. <BR>Saving changes "
                            "to this form without validation may lead to serious data "
                            "integrity errors. <BR><B>If you did not expect this, contact "
                            "your data manager immediately and do not continue.</B>"
                        ),  # nosec B703, B308
                    ),
                )
            except MessageFailure:
                pass
            form_cls = self.custom_modelform_factory()
        else:
            form_cls = self.form
        return form_cls

    def get_form(self, request, obj=None, change=False, **kwargs):
        """
        Return a Form class for use in the admin add view. This is used by
        add_view and change_view.

        EDC Note:
            The only change to this method is `get_form_cls` sets the
            `form_cls` variable instead of calling `self.form` throughout.
        """
        form_cls = self.get_form_cls(request)

        if "fields" in kwargs:
            fields = kwargs.pop("fields")
        else:
            fields = flatten_fieldsets(self.get_fieldsets(request, obj))
        excluded = self.get_exclude(request, obj)
        exclude = [] if excluded is None else list(excluded)
        readonly_fields = self.get_readonly_fields(request, obj)
        exclude.extend(readonly_fields)
        # Exclude all fields if it's a change form and the user doesn't have
        # the change permission.
        if (
            change
            and hasattr(request, "user")
            and not self.has_change_permission(request, obj)
        ):
            exclude.extend(fields)
        if excluded is None and hasattr(form_cls, "_meta") and form_cls._meta.exclude:
            # Take the custom ModelForm's Meta.exclude into account only if the
            # ModelAdmin doesn't define its own.
            exclude.extend(form_cls._meta.exclude)
        # if exclude is an empty list we pass None to be consistent with the
        # default on modelform_factory
        exclude = exclude or None

        # Remove declared form fields which are in readonly_fields.
        new_attrs = dict.fromkeys(f for f in readonly_fields if f in self.form.declared_fields)
        form = type(form_cls.__name__, (form_cls,), new_attrs)

        defaults = {
            "form": form,
            "fields": fields,
            "exclude": exclude,
            "formfield_callback": partial(self.formfield_for_dbfield, request=request),
            **kwargs,
        }

        if defaults["fields"] is None and not modelform_defines_fields(defaults["form"]):
            defaults["fields"] = forms.ALL_FIELDS

        try:
            return modelform_factory(self.model, **defaults)
        except FieldError as e:
            raise FieldError(
                "%s. Check fields/fieldsets/exclude attributes of class %s."
                % (e, self.__class__.__name__)
            )



from __future__ import annotations

from datetime import date, datetime

from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist

from edc_constants.constants import YES
from edc_dx_review.utils import (
    get_clinical_review_baseline_model_cls,
    get_clinical_review_model_cls,
    get_initial_review_model_cls,
)

from .utils import get_diagnosis_labels


class InitialReviewRequired(Exception):
    pass


class MultipleInitialReviewsExist(Exception):
    pass


class ClinicalReviewBaselineRequired(Exception):
    pass


class DiagnosesError(Exception):
    pass


class Diagnoses:
    """
    Tightly coupled to models
    """

    def __init__(
        self,
        subject_identifier: str = None,
        report_datetime: datetime = None,
        subject_visit=None,
        lte: bool | None = None,
        limit_to_single_condition_prefix: str | None = None,
    ) -> None:
        self.single_condition_prefix = (
            limit_to_single_condition_prefix.lower()
            if limit_to_single_condition_prefix
            else None
        )
        if subject_visit:
            if subject_identifier or report_datetime:
                raise DiagnosesError(
                    "Ambiguous parameters provided. Expected either "
                    "`subject_visit` or `subject_identifier, report_datetime`. Not both."
                )
            self.report_datetime = subject_visit.report_datetime
            self.subject_identifier = subject_visit.appointment.subject_identifier
        else:
            self.report_datetime = report_datetime
            self.subject_identifier = subject_identifier
        self.lte = lte
        self.clinical_review_baseline_exists_or_raise()

    @property
    def diagnosis_labels(self):
        if self.single_condition_prefix:
            return {
                k.lower(): v
                for k, v in get_diagnosis_labels().items()
                if k == self.single_condition_prefix
            }
        return get_diagnosis_labels()

    def get_dx_by_model(self, instance) -> str:
        dx = None
        for prefix in self.diagnosis_labels:
            if instance.__class__.__name__.lower().startswith(prefix.lower()):
                dx = self.get_dx(prefix)
                break
        if not dx:
            raise DiagnosesError(
                f"Invalid. No diagnoses detected. "
                f"See responses on {self.clinical_review_baseline._meta.verbose_name}."
            )
        return dx

    def get_dx_date(self, prefix: str) -> date | None:
        """Returns a dx date from the initial review for the condition.

        Raises if initial review does not exist."""
        prefix = prefix.lower()
        if self.initial_reviews.get(prefix):
            return self.initial_reviews.get(prefix).get_best_dx_date()
        return None

    def get_dx(self, prefix: str) -> str | None:
        """Returns YES if any diagnoses for this condition otherwise None.

        References clinical_review_baseline, clinical_review

        name is `dm`, `hiv` or `htn`.
        """
        diagnoses = [
            getattr(self.clinical_review_baseline, f"{prefix.lower()}_dx", "") == YES,
            *[
                (getattr(obj, f"{prefix.lower()}_dx", "") == YES)
                for obj in self.clinical_reviews
            ],
        ]
        if any(diagnoses):
            return YES
        return None

    def clinical_review_baseline_exists_or_raise(self):
        return self.clinical_review_baseline

    @property
    def clinical_review_baseline(self):
        try:
            obj = get_clinical_review_baseline_model_cls().objects.get(
                subject_visit__subject_identifier=self.subject_identifier,
            )
        except ObjectDoesNotExist:
            raise ClinicalReviewBaselineRequired(
                "Please complete "
                f"{get_clinical_review_baseline_model_cls()._meta.verbose_name}."
            )
        return obj

    def report_datetime_opts(
        self, prefix: str = None, lte: bool = None
    ) -> dict[str, datetime]:
        opts = {}
        prefix = prefix.lower() or ""
        if self.report_datetime:
            if lte or self.lte:
                opts.update({f"{prefix.lower()}report_datetime__lte": self.report_datetime})
            else:
                opts.update({f"{prefix.lower()}report_datetime__lt": self.report_datetime})
        return opts

    @property
    def clinical_reviews(self):
        return get_clinical_review_model_cls().objects.filter(
            subject_visit__subject_identifier=self.subject_identifier,
            **self.report_datetime_opts("subject_visit__"),
        )

    def get_initial_reviews(self):
        return self.initial_reviews

    @property
    def initial_reviews(self):
        """Returns a dict of initial review model instances
        for each diagnosis.

        If any initial review is expected but does not exist,
        an expection is raised.
        """
        initial_reviews = {}

        options = []
        for prefix, label in self.diagnosis_labels.items():
            prefix = prefix.lower()
            options.append(
                (
                    prefix,
                    self.get_dx(prefix),
                    get_initial_review_model_cls(prefix),
                    f"{label.title()} diagnosis",
                )
            )
        for name, diagnosis, initial_review_model_cls, description in options:
            if diagnosis:
                extra_msg = description.title()
                try:
                    obj = initial_review_model_cls.objects.get(
                        subject_visit__subject_identifier=self.subject_identifier,
                        **self.report_datetime_opts("subject_visit__", lte=True),
                    )
                except ObjectDoesNotExist:
                    subject_visit = self.initial_diagnosis_visit(name)
                    if subject_visit:
                        visit_label = (
                            f"{subject_visit.visit_code}."
                            f"{subject_visit.visit_code_sequence}"
                        )
                        extra_msg = f"{description} was reported on visit {visit_label}. "
                    raise InitialReviewRequired(
                        f"{extra_msg}. Complete the "
                        f"`{initial_review_model_cls._meta.verbose_name}` CRF first."
                    )
                except MultipleObjectsReturned:
                    qs = initial_review_model_cls.objects.filter(
                        subject_visit__subject_identifier=self.subject_identifier,
                        **self.report_datetime_opts("subject_visit__", lte=True),
                    ).order_by(
                        "subject_visit__visit_code",
                        "subject_visit__visit_code_sequence",
                    )
                    visits_str = ", ".join(
                        [
                            (
                                f"{obj.subject_visit.visit_code}."
                                f"{obj.subject_visit.visit_code_sequence}"
                            )
                            for obj in qs
                        ]
                    )
                    raise MultipleInitialReviewsExist(
                        f"More than one `{initial_review_model_cls._meta.verbose_name}` "
                        f"has been submitted. "
                        f"This needs to be corrected. Try removing all but the first "
                        f"`{initial_review_model_cls._meta.verbose_name}` "
                        "before continuing. "
                        f"`{initial_review_model_cls._meta.verbose_name}` "
                        "CRFs have been submitted "
                        f"for visits {visits_str}"
                    )

                else:
                    initial_reviews.update({name: obj})
        return initial_reviews

    def initial_diagnosis_visit(self, prefix):
        related_visit_model_attr = (
            get_clinical_review_baseline_model_cls().related_visit_model_attr()
        )
        opts = {
            f"{related_visit_model_attr}__subject_identifier": self.subject_identifier,
            f"{prefix.lower()}_dx": YES,
        }
        opts.update(**self.report_datetime_opts(f"{related_visit_model_attr}__", lte=True))
        try:
            clinical_review_baseline = get_clinical_review_baseline_model_cls().objects.get(
                **opts
            )
        except ObjectDoesNotExist:
            subject_visit = None
        else:
            subject_visit = clinical_review_baseline.related_visit
        if not subject_visit:
            try:
                clinical_review = get_clinical_review_model_cls().objects.get(**opts)
            except ObjectDoesNotExist:
                subject_visit = None
            else:
                subject_visit = clinical_review.related_visit
        return subject_visit



import contextlib
from typing import Any


def calculate_missing(obj: Any, panel: Any) -> tuple[int, str | None]:
    """Returns a tuple of (missing_count, missing)

    Counts the number of blank results within the model
    instance (missing_count).

    Creates a string of utest_ids of the missing field values
    (missing).

    Only for model classes whose field classes follow the naming
    convention tied to a lab panel.

    See also: BloodResultsMethodsModelMixin and
    BloodResultsFieldsModelMixin.
    """
    fields = [fld_cls.name for fld_cls in obj._meta.get_fields()]
    missing = []
    for utest_id in panel.utest_ids:
        for field in fields:
            with contextlib.suppress(ValueError):
                utest_id, _ = utest_id  # noqa: PLW2901
            if field == utest_id or (
                field == f"{utest_id}_value" and getattr(obj, field) is None
            ):
                missing.append(field)
    return len(missing), ",".join(missing) if missing else ""



# Generated by Django 4.2.4 on 2023-08-23 01:13

import _socket
import django_audit_fields.fields.hostname_modification_field
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_action_item", "0033_alter_actionitem_managers"),
    ]

    operations = [
        migrations.AlterField(
            model_name="actionitem",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="actionitem",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="actiontype",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="actiontype",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="actiontype",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="actiontype",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="edcpermissions",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="edcpermissions",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="edcpermissions",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="edcpermissions",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalactionitem",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreference",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="historicalreference",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="historicalreference",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="historicalreference",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="device_created",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device created"),
        ),
        migrations.AlterField(
            model_name="reference",
            name="device_modified",
            field=models.CharField(blank=True, max_length=10, verbose_name="Device modified"),
        ),
        migrations.AlterField(
            model_name="reference",
            name="hostname_created",
            field=models.CharField(
                blank=True,
                default=_socket.gethostname,
                help_text="System field. (modified on create only)",
                max_length=60,
                verbose_name="Hostname created",
            ),
        ),
        migrations.AlterField(
            model_name="reference",
            name="hostname_modified",
            field=django_audit_fields.fields.hostname_modification_field.HostnameModificationField(
                blank=True,
                help_text="System field. (modified on every save)",
                max_length=50,
                verbose_name="Hostname modified",
            ),
        ),
    ]



from django import forms

from ..modelform_mixins import AeModelFormMixin, AeTmgModelFormMixin
from ..utils import get_ae_model


class AeTmgForm(AeTmgModelFormMixin, AeModelFormMixin, forms.ModelForm):
    class Meta(AeTmgModelFormMixin.Meta):
        model = get_ae_model("aetmg")
        fields = "__all__"
